#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# pylint: disable=too-many-lines
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import dateutil
import pandas as pd
import trafaret as t

from datarobot.enums import (
    BUCKET_SIZE,
    CustomMetricAggregationType,
    CustomMetricBucketTimeStep,
    CustomMetricDirectionality,
    DEFAULT_MAX_WAIT,
    HostedCustomMetricsTemplateMetricTypeQueryParams,
    ListHostedCustomMetricTemplatesSortQueryParams,
)
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.models.registry import Job
from datarobot.models.runtime_parameters import RuntimeParameterValue
from datarobot.models.types import Schedule
from datarobot.utils import from_api, to_api, underscorize
from datarobot.utils.waiters import wait_for_async_resolution

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class BaselineValues(TypedDict, total=False):
        value: float

    class DatasetColumn(TypedDict, total=False):
        column_name: str
        time_format: Optional[str]

    class CustomMetricBucket(TypedDict):
        value: float
        sample_size: int
        timestamp: Optional[Union[datetime, str]]
        batch: Optional[str]
        association_id: Optional[str]

    class CustomMetricSegmentFromJSON(TypedDict):
        name: str
        value: str

    class CustomMetricSegmentFromDataset(TypedDict):
        name: str
        column: str

    class Period(TypedDict, total=False):
        start: datetime
        end: datetime

    class Summary(TypedDict, total=False):
        period: Period

    class Bucket(TypedDict):
        period: Period
        value: int
        sample_size: int

    class Batch(TypedDict, total=False):
        id: str
        name: str
        created_at: datetime
        last_prediction_timestamp: datetime

    class BatchBucket(TypedDict):
        batch: Batch
        value: int
        sample_size: int


class CustomMetric(APIObject):
    """A DataRobot custom metric.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the custom metric.
    deployment_id: str
        The ID of the deployment.
    name: str
        The name of the custom metric.
    units: str
        The units, or the y-axis label, of the given custom metric.
    baseline_values: BaselinesValues
        The baseline value used to add "reference dots" to the values over time chart.
    is_model_specific: bool
        Determines whether the metric is related to the model or deployment.
    type: CustomMetricAggregationType
        The aggregation type of the custom metric.
    directionality: CustomMetricDirectionality
        The directionality of the custom metric.
    time_step: CustomMetricBucketTimeStep
        Custom metric time bucket size.
    description: str
        A description of the custom metric.
    association_id: DatasetColumn
        A custom metric association_id column source when reading values from columnar dataset.
    timestamp: DatasetColumn
        A custom metric timestamp column source when reading values from columnar dataset.
    value: DatasetColumn
        A custom metric value source when reading values from columnar dataset.
    sample_count: DatasetColumn
        A custom metric sample source when reading values from columnar dataset.
    batch: str
        A custom metric batch ID source when reading values from columnar dataset.
    """

    _path = "deployments/{}/customMetrics/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("units"): t.String(),
            t.Key("baseline_values"): t.List(t.Dict().allow_extra("*")),
            t.Key("is_model_specific"): t.Bool(),
            t.Key("type"): t.String(),
            t.Key("directionality"): t.String(),
            t.Key("time_step"): t.String(),
            t.Key("description", optional=True): t.Or(t.String(allow_blank=True), t.Null),
            t.Key("association_id", optional=True): t.Dict().allow_extra("*"),
            t.Key("value", optional=True): t.Dict().allow_extra("*"),
            t.Key("sample_count", optional=True): t.Dict().allow_extra("*"),
            t.Key("timestamp", optional=True): t.Dict().allow_extra("*"),
            t.Key("batch", optional=True): t.Dict().allow_extra("*"),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        units: str,
        baseline_values: BaselineValues,
        is_model_specific: bool,
        type: CustomMetricAggregationType,
        directionality: CustomMetricDirectionality,
        time_step: str = CustomMetricBucketTimeStep.HOUR,
        description: Optional[str] = None,
        association_id: Optional[DatasetColumn] = None,
        value: Optional[DatasetColumn] = None,
        sample_count: Optional[DatasetColumn] = None,
        timestamp: Optional[DatasetColumn] = None,
        batch: Optional[DatasetColumn] = None,
        deployment_id: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.units = units
        self.baseline_values = baseline_values
        self.is_model_specific = is_model_specific
        self.type = type
        self.directionality = directionality
        self.time_step = time_step
        self.description = description
        self.association_id = association_id
        self.value = value
        self.sample_count = sample_count
        self.timestamp = timestamp
        self.batch = batch
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "CustomMetric({} | {})".format(self.id, self.name)

    @classmethod
    def create(
        cls,
        name: str,
        deployment_id: str,
        units: str,
        is_model_specific: bool,
        aggregation_type: CustomMetricAggregationType,
        directionality: CustomMetricDirectionality,
        time_step: str = CustomMetricBucketTimeStep.HOUR,
        description: Optional[str] = None,
        baseline_value: Optional[float] = None,
        value_column_name: Optional[str] = None,
        sample_count_column_name: Optional[str] = None,
        timestamp_column_name: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        batch_column_name: Optional[str] = None,
    ) -> CustomMetric:
        """Create a custom metric for a deployment

        Parameters
        ----------
        name: str
            The name of the custom metric.
        deployment_id: str
            The id of the deployment.
        units: str
            The units, or the y-axis label, of the given custom metric.
        baseline_value: float
            The baseline value used to add "reference dots" to the values over time chart.
        is_model_specific: bool
            Determines whether the metric is related to the model or deployment.
        aggregation_type: CustomMetricAggregationType
            The aggregation type of the custom metric.
        directionality: CustomMetricDirectionality
            The directionality of the custom metric.
        time_step: CustomMetricBucketTimeStep
            Custom metric time bucket size.
        description: Optional[str]
            A description of the custom metric.
        value_column_name: Optional[str]
            A custom metric value column name when reading values from columnar dataset.
        sample_count_column_name: Optional[str]
            Points to a weight column name if users provide pre-aggregated metric values from columnar dataset.
        timestamp_column_name: Optional[str]
            A custom metric timestamp column name when reading values from columnar dataset.
        timestamp_format: Optional[str]
            A custom metric timestamp format when reading values from columnar dataset.
        batch_column_name: Optional[str]
            A custom metric batch ID column name when reading values from columnar dataset.

        Returns
        -------
        CustomMetric
            The custom metric object.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric
            from datarobot.enums import CustomMetricAggregationType, CustomMetricDirectionality

            custom_metric = CustomMetric.create(
                deployment_id="5c939e08962d741e34f609f0",
                name="Sample metric",
                units="Y",
                baseline_value=12,
                is_model_specific=True,
                aggregation_type=CustomMetricAggregationType.AVERAGE,
                directionality=CustomMetricDirectionality.HIGHER_IS_BETTER
                )
        """
        payload = {
            "name": name,
            "units": units,
            "baselineValues": [{"value": baseline_value}] if baseline_value else [],
            "isModelSpecific": is_model_specific,
            "type": aggregation_type,
            "directionality": directionality,
            "timeStep": time_step,
            "description": description if description else "",
        }
        if value_column_name is not None:
            payload["value"] = {"columnName": value_column_name}
        if sample_count_column_name is not None:
            payload["sampleCount"] = {"columnName": sample_count_column_name}
        if timestamp_column_name is not None:
            payload["timestamp"] = {
                "columnName": timestamp_column_name,
                "timeFormat": timestamp_format,
            }
        if batch_column_name is not None:
            payload["batch"] = {"columnName": batch_column_name}
        path = cls._path.format(deployment_id)
        response = cls._client.post(path, json=payload)
        custom_metric_id = response.json()["id"]
        return cls.get(deployment_id, custom_metric_id)

    @classmethod
    def get(cls, deployment_id: str, custom_metric_id: str) -> CustomMetric:
        """Get a custom metric for a deployment

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        custom_metric_id: str
            The ID of the custom metric.

        Returns
        -------
        CustomMetric
            The custom metric object.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )

            custom_metric.id
            >>>'65f17bdcd2d66683cdfc1113'
        """
        path = "{}{}/".format(cls._path.format(deployment_id), custom_metric_id)
        custom_metric = cls.from_location(path)
        custom_metric.deployment_id = deployment_id
        return custom_metric

    @classmethod
    def list(cls, deployment_id: str) -> List[CustomMetric]:
        """List all custom metrics for a deployment

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.

        Returns
        -------
        custom_metrics: list
            A list of custom metrics objects.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metrics = CustomMetric.list(deployment_id="5c939e08962d741e34f609f0")
            custom_metrics[0].id
            >>>'65f17bdcd2d66683cdfc1113'
        """
        path = cls._path.format(deployment_id)
        get_response = cls._client.get(path).json()
        custom_metrics = [cls.from_server_data(data) for data in get_response["data"]]
        for custom_metric in custom_metrics:
            custom_metric.deployment_id = deployment_id
        return custom_metrics

    @classmethod
    def delete(cls, deployment_id: str, custom_metric_id: str) -> None:
        """Delete a custom metric associated with a deployment.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        custom_metric_id: str
            The ID of the custom metric.

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            CustomMetric.delete(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
        """
        path = "{}{}/".format(cls._path.format(deployment_id), custom_metric_id)
        cls._client.delete(path)

    def update(
        self,
        name: Optional[str] = None,
        units: Optional[str] = None,
        aggregation_type: Optional[CustomMetricAggregationType] = None,
        directionality: Optional[CustomMetricDirectionality] = None,
        time_step: Optional[str] = None,
        description: Optional[str] = None,
        baseline_value: Optional[float] = None,
        value_column_name: Optional[str] = None,
        sample_count_column_name: Optional[str] = None,
        timestamp_column_name: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        batch_column_name: Optional[str] = None,
    ) -> CustomMetric:
        """Update metadata of a custom metric

        Parameters
        ----------
        name: Optional[str]
            The name of the custom metric.
        units: Optional[str]
            The units, or the y-axis label, of the given custom metric.
        baseline_value: Optional[float]
            The baseline value used to add "reference dots" to the values over time chart.
        aggregation_type: Optional[CustomMetricAggregationType]
            The aggregation type of the custom metric.
        directionality: Optional[CustomMetricDirectionality]
            The directionality of the custom metric.
        time_step: Optional[CustomMetricBucketTimeStep]
            Custom metric time bucket size.
        description: Optional[str]
            A description of the custom metric.
        value_column_name: Optional[str]
            A custom metric value column name when reading values from columnar dataset.
        sample_count_column_name: Optional[str]
            Points to a weight column name if users provide pre-aggregated metric values from columnar dataset.
        timestamp_column_name: Optional[str]
            A custom metric timestamp column name when reading values from columnar dataset.
        timestamp_format: Optional[str]
            A custom metric timestamp format when reading values from columnar dataset.
        batch_column_name: Optional[str]
            A custom metric batch ID column name when reading values from columnar dataset.

        Returns
        -------
        CustomMetric
            The custom metric object.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric
            from datarobot.enums import CustomMetricAggregationType, CustomMetricDirectionality

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            custom_metric = custom_metric.update(
                deployment_id="5c939e08962d741e34f609f0",
                name="Sample metric",
                units="Y",
                baseline_value=12,
                is_model_specific=True,
                aggregation_type=CustomMetricAggregationType.AVERAGE,
                directionality=CustomMetricDirectionality.HIGHER_IS_BETTER
                )
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if units is not None:
            params["units"] = units
        if aggregation_type is not None:
            params["type"] = aggregation_type
        if directionality is not None:
            params["directionality"] = directionality
        if baseline_value is not None:
            params["baselineValues"] = [{"value": baseline_value}]
        if time_step is not None:
            params["timeStep"] = time_step
        if description is not None:
            params["description"] = description
        if value_column_name is not None:
            params["value"] = {"columnName": value_column_name}
        if sample_count_column_name is not None:
            params["sampleCount"] = {"columnName": sample_count_column_name}
        if timestamp_column_name is not None:
            params["timestamp"] = {
                "columnName": timestamp_column_name,
                "timeFormat": timestamp_format,
            }
        if batch_column_name is not None:
            params["batch"] = {"columnName": batch_column_name}

        path = "{}{}/".format(self._path.format(self.deployment_id), self.id)
        self._client.patch(path, data=params)

        for param, value in params.items():
            case_converted = from_api(value)
            param_converted = underscorize(param)
            setattr(self, param_converted, case_converted)
        return self

    def unset_baseline(self) -> None:
        """Unset the baseline value of a custom metric

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric
            from datarobot.enums import CustomMetricAggregationType, CustomMetricDirectionality

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            custom_metric.baseline_values
            >>> [{'value': 12.0}]
            custom_metric.unset_baseline()
            custom_metric.baseline_values
            >>> []
        """
        params: Dict[str, Any] = {"baselineValues": []}
        path = "{}{}/".format(self._path.format(self.deployment_id), self.id)
        self._client.patch(path, data=params)
        for param, value in params.items():
            case_converted = from_api(value)
            param_converted = underscorize(param)
            setattr(self, param_converted, case_converted)

    def submit_values(
        self,
        data: Union[pd.DataFrame, List[CustomMetricBucket]],
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        dry_run: Optional[bool] = False,
        segments: Optional[List[CustomMetricSegmentFromJSON]] = None,
    ) -> None:
        """Submit aggregated custom metrics values from JSON.

        Parameters
        ----------
        data: pd.DataFrame or List[CustomMetricBucket]
            The data containing aggregated custom metric values.
        model_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model is not needed.
        model_package_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model package is not needed.
        dry_run: Optional[bool]
            Specifies whether or not metric data is submitted in production mode (where data is saved).
        segments: Optional[CustomMetricSegmentFromJSON]
            A list of segments for a custom metric used in segmented analysis.

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )

            # data for values over time
            data = [{
                'value': 12,
                'sample_size': 3,
                'timestamp': '2024-03-15T14:00:00'
            }]

            # data witch association ID
            data = [{
                'value': 12,
                'sample_size': 3,
                'timestamp': '2024-03-15T14:00:00',
                'association_id': '65f44d04dbe192b552e752ed'
            }]

            # data for batches
            data = [{
                'value': 12,
                'sample_size': 3,
                'batch': '65f44c93fedc5de16b673a0d'
            }]

            # for deployment specific metrics
            custom_metric.submit_values(data=data)

            # for model specific metrics pass model_package_id or model_id
            custom_metric.submit_values(data=data, model_package_id="6421df32525c58cc6f991f25")

            # dry run
            custom_metric.submit_values(data=data, model_package_id="6421df32525c58cc6f991f25", dry_run=True)

            # for segmented analysis
            segments = [{"name": "custom_seg", "value": "val_1"}]
            custom_metric.submit_values(data=data, model_package_id="6421df32525c58cc6f991f25", segments=segments)
        """
        if not isinstance(data, (list, pd.DataFrame)):
            raise ValueError(
                "data should be either a list of dict-like objects or a pandas.DataFrame"
            )

        if isinstance(data, pd.DataFrame):
            data = data.to_dict(orient="records")

        buckets = []
        for row in data:
            bucket = {"sampleSize": row["sample_size"], "value": row["value"]}

            if "timestamp" in row and "batch" in row:
                raise ValueError("data should contain either timestamps or batch IDs")

            if "timestamp" in row:
                if isinstance(row["timestamp"], datetime):
                    bucket["timestamp"] = row["timestamp"].isoformat()
                else:
                    bucket["timestamp"] = dateutil.parser.parse(row["timestamp"]).isoformat()

            if "batch" in row:
                bucket["batch"] = row["batch"]

            if "association_id" in row:
                bucket["associationId"] = row["association_id"]

            buckets.append(bucket)

        payload: Dict[str, Any] = {"buckets": buckets, "dryRun": dry_run}
        if model_id:
            payload["modelId"] = model_id
        if model_package_id:
            payload["modelPackageId"] = model_package_id
        if segments:
            payload["segments"] = segments

        path = "{}{}/fromJSON/".format(self._path.format(self.deployment_id), self.id)
        response = self._client.post(path, data=payload)
        if not dry_run:
            wait_for_async_resolution(self._client, response.headers["Location"], DEFAULT_MAX_WAIT)

    def submit_single_value(
        self,
        value: float,
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        dry_run: Optional[bool] = False,
        segments: Optional[List[CustomMetricSegmentFromJSON]] = None,
    ) -> None:
        """Submit a single custom metric value at the current moment.

        Parameters
        ----------
        value: float
            Single numeric custom metric value.
        model_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model is not needed.
        model_package_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model package is not needed.
        dry_run: Optional[bool]
            Specifies whether or not metric data is submitted in production mode (where data is saved).
        segments: Optional[CustomMetricSegmentFromJSON]
            A list of segments for a custom metric used in segmented analysis.

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )

            # for deployment specific metrics
            custom_metric.submit_single_value(value=121)

            # for model specific metrics pass model_package_id or model_id
            custom_metric.submit_single_value(value=121, model_package_id="6421df32525c58cc6f991f25")

            # dry run
            custom_metric.submit_single_value(value=121, model_package_id="6421df32525c58cc6f991f25", dry_run=True)

            # for segmented analysis
            segments = [{"name": "custom_seg", "value": "val_1"}]
            custom_metric.submit_single_value(value=121, model_package_id="6421df32525c58cc6f991f25", segments=segments)
        """
        bucket = {"sampleSize": 1, "value": value, "timestamp": datetime.now().isoformat()}
        payload: Dict[str, Any] = {"buckets": [bucket], "dryRun": dry_run}
        if model_id:
            payload["modelId"] = model_id
        if model_package_id:
            payload["modelPackageId"] = model_package_id
        if segments:
            payload["segments"] = segments

        path = "{}{}/fromJSON/".format(self._path.format(self.deployment_id), self.id)
        response = self._client.post(path, data=payload)
        if not dry_run:
            wait_for_async_resolution(self._client, response.headers["Location"], DEFAULT_MAX_WAIT)

    def submit_values_from_catalog(
        self,
        dataset_id: str,
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        batch_id: Optional[str] = None,
        segments: Optional[List[CustomMetricSegmentFromDataset]] = None,
    ) -> None:
        """Submit aggregated custom metrics values from dataset (AI catalog).
        The names of the columns in the dataset should correspond to the names of the columns that were defined in
        the custom metric. In addition, the format of the timestamps should also be the same as defined in the metric.

        Parameters
        ----------
        dataset_id: str
            The ID of the source dataset.
        model_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model is not needed.
        model_package_id: Optional[str]
            For a model metric: the ID of the associated champion/challenger model, used to update the metric values.
            For a deployment metric: the ID of the model package is not needed.
        batch_id: Optional[str]
            Specifies a batch ID associated with all values provided by this dataset, an alternative
            to providing batch IDs as a column within a dataset (at the record level).
        segments: Optional[CustomMetricSegmentFromDataset]
            A list of segments for a custom metric used in segmented analysis.

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )

            # for deployment specific metrics
            custom_metric.submit_values_from_catalog(dataset_id="61093144cabd630828bca321")

            # for model specific metrics pass model_package_id or model_id
            custom_metric.submit_values_from_catalog(
                dataset_id="61093144cabd630828bca321",
                model_package_id="6421df32525c58cc6f991f25"
            )

            # for segmented analysis
            segments = [{"name": "custom_seg", "column": "column_with_segment_values"}]
            custom_metric.submit_values_from_catalog(
                dataset_id="61093144cabd630828bca321",
                model_package_id="6421df32525c58cc6f991f25",
                segments=segments
            )
        """
        payload: Dict[str, Any] = {
            "datasetId": dataset_id,
            "value": self.value,
            "timestamp": self.timestamp,
            "sampleCount": self.sample_count,
            "batch": self.batch,
            "associationId": self.association_id,
        }

        if model_id:
            payload["modelId"] = model_id
        if model_package_id:
            payload["modelPackageId"] = model_package_id
        if segments:
            payload["segments"] = segments
        if batch_id:
            payload["batchId"] = batch_id

        path = "{}{}/fromDataset/".format(self._path.format(self.deployment_id), self.id)
        response = self._client.post(path, data=to_api(payload))
        wait_for_async_resolution(self._client, response.headers["Location"], DEFAULT_MAX_WAIT)

    def get_values_over_time(
        self,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_package_id: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
        bucket_size: str = BUCKET_SIZE.P7D,
    ) -> CustomMetricValuesOverTime:
        """Retrieve values of a single custom metric over a time period.

        Parameters
        ----------
        start: datetime or str
            Start of the time period.
        end: datetime or str
            End of the time period.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        bucket_size: Optional[str]
            Time duration of a bucket, in ISO 8601 time duration format.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_over_time: CustomMetricValuesOverTime
            The queried custom metric values over time information.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric
            from datetime import datetime, timedelta

            now=datetime.now()
            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            values_over_time = custom_metric.get_values_over_time(start=now - timedelta(days=7), end=now)

            values_over_time.bucket_values
            >>> {datetime.datetime(2024, 3, 22, 14, 0, tzinfo=tzutc()): 1.0,
            >>> datetime.datetime(2024, 3, 22, 15, 0, tzinfo=tzutc()): 123.0}}

            values_over_time.bucket_sample_sizes
            >>> {datetime.datetime(2024, 3, 22, 14, 0, tzinfo=tzutc()): 1,
            >>>  datetime.datetime(2024, 3, 22, 15, 0, tzinfo=tzutc()): 1}}

            values_over_time.get_buckets_as_dataframe()
            >>>                        start                       end  value  sample_size
            >>> 0  2024-03-21 16:00:00+00:00 2024-03-21 17:00:00+00:00    NaN          NaN
            >>> 1  2024-03-21 17:00:00+00:00 2024-03-21 18:00:00+00:00    NaN          NaN
        """

        if not self.deployment_id:
            raise ValueError("Deployment ID is required to get custom metric values over time.")

        return CustomMetricValuesOverTime.get(
            custom_metric_id=self.id,
            deployment_id=self.deployment_id,
            start=start,
            end=end,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
            bucket_size=bucket_size,
        )

    def get_summary(
        self,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_package_id: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricSummary:
        """Retrieve the summary of a custom metric over a time period.

        Parameters
        ----------
        start: datetime or str
            Start of the time period.
        end: datetime or str
            End of the time period.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_summary: CustomMetricSummary
            The summary of the custom metric.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric
            from datetime import datetime, timedelta

            now=datetime.now()
            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            summary = custom_metric.get_summary(start=now - timedelta(days=7), end=now)

            print(summary)
            >> "CustomMetricSummary(2024-03-21 15:52:13.392178+00:00 - 2024-03-22 15:52:13.392168+00:00:
            {'id': '65fd9b1c0c1a840bc6751ce0', 'name': 'Test METRIC', 'value': 215.0, 'sample_count': 13,
            'baseline_value': 12.0, 'percent_change': 24.02})"
        """

        if not self.deployment_id:
            raise ValueError("Deployment ID is required to get custom metric summary.")

        return CustomMetricSummary.get(
            custom_metric_id=self.id,
            deployment_id=self.deployment_id,
            start=start,
            end=end,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )

    def get_values_over_batch(
        self,
        batch_ids: Optional[List[str]] = None,
        model_package_id: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricValuesOverBatch:
        """Retrieve values of a single custom metric over batches.

        Parameters
        ----------
        batch_ids : Optional[List[str]]
            Specify a list of batch IDs to pull the data for.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_over_batch: CustomMetricValuesOverBatch
            The queried custom metric values over batch information.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            # all batch metrics all model specific
            values_over_batch = custom_metric.get_values_over_batch(model_package_id='6421df32525c58cc6f991f25')

            values_over_batch.bucket_values
            >>> {'6572db2c9f9d4ad3b9de33d0': 35.0, '6572db2c9f9d4ad3b9de44e1': 105.0}

            values_over_batch.bucket_sample_sizes
            >>> {'6572db2c9f9d4ad3b9de33d0': 6, '6572db2c9f9d4ad3b9de44e1': 8}


            values_over_batch.get_buckets_as_dataframe()
            >>>                    batch_id                     batch_name  value  sample_size
            >>> 0  6572db2c9f9d4ad3b9de33d0  Batch 1 - 03/26/2024 13:04:46   35.0            6
            >>> 1  6572db2c9f9d4ad3b9de44e1  Batch 2 - 03/26/2024 13:06:04  105.0            8
        """

        if not self.deployment_id:
            raise ValueError("Deployment ID is required to get custom metric values over time.")

        if not model_package_id and not model_id:
            raise ValueError(
                "For batch metrics either the modelPackageId or the modelId must be passed."
            )

        return CustomMetricValuesOverBatch.get(
            custom_metric_id=self.id,
            deployment_id=self.deployment_id,
            batch_ids=batch_ids,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )

    def get_batch_summary(
        self,
        batch_ids: Optional[List[str]] = None,
        model_package_id: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricBatchSummary:
        """Retrieve the summary of a custom metric over a batch.

        Parameters
        ----------
        batch_ids : Optional[List[str]]
            Specify a list of batch IDs to pull the data for.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_summary: CustomMetricBatchSummary
            The batch summary of the custom metric.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import CustomMetric

            custom_metric = CustomMetric.get(
                deployment_id="5c939e08962d741e34f609f0",
                custom_metric_id="65f17bdcd2d66683cdfc1113"
            )
            # all batch metrics all model specific
            batch_summary = custom_metric.get_batch_summary(model_package_id='6421df32525c58cc6f991f25')

            print(batch_summary)
            >> CustomMetricBatchSummary({'id': '6605396413434b3a7b74342c', 'name': 'batch metric', 'value': 41.25,
            'sample_count': 28, 'baseline_value': 123.0, 'percent_change': -66.46})
        """

        if not self.deployment_id:
            raise ValueError("Deployment ID is required to get custom metric values over time.")

        if not model_package_id and not model_id:
            raise ValueError(
                "For batch metrics either the modelPackageId or the modelId must be passed."
            )

        return CustomMetricBatchSummary.get(
            custom_metric_id=self.id,
            deployment_id=self.deployment_id,
            batch_ids=batch_ids,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )


class CustomMetricValuesOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Custom metric over time information.

    .. versionadded:: v3.4

    Attributes
    ----------
    buckets: List[Bucket]
        A list of bucketed time periods and the custom metric values aggregated over that period.
    summary: Summary
        The summary of values over time retrieval.
    metric: Dict
        A custom metric definition.
    deployment_id: str
        The ID of the deployment.
    segment_attribute: str
        The name of the segment on which segment analysis is being performed.
    segment_value: str
        The value of the segment_attribute to segment on.
    """

    _path = "deployments/{}/customMetrics/{}/valuesOverTime/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> dateutil.parser.parse,
            t.Key("end"): t.String >> dateutil.parser.parse,
        }
    )
    _bucket = t.Dict(
        {
            t.Key("period"): t.Or(_period, t.Null),
            t.Key("value"): t.Or(t.Float, t.Null),
            t.Key("sample_size"): t.Or(t.Int, t.Null),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("buckets"): t.List(_bucket),
            t.Key("summary"): t.Dict(
                {
                    t.Key("start"): t.String >> dateutil.parser.parse,
                    t.Key("end"): t.String >> dateutil.parser.parse,
                }
            ),
            t.Key("metric"): t.Dict().allow_extra("*"),
            t.Key("segment_attribute", optional=True): t.String(),
            t.Key("segment_value", optional=True): t.String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        buckets: Optional[List[Bucket]] = None,
        summary: Optional[Summary] = None,
        metric: Optional[Dict[str, Any]] = None,
        deployment_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ):
        self.buckets = buckets if buckets is not None else []
        self.summary = summary if summary is not None else {}
        self.metric = metric
        self.deployment_id = deployment_id
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        return "CustomMetricValuesOverTime({} - {})".format(
            self.summary.get("start"),
            self.summary.get("end"),
        )

    @classmethod
    def get(
        cls,
        deployment_id: str,
        custom_metric_id: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
        bucket_size: str = BUCKET_SIZE.P7D,
    ) -> CustomMetricValuesOverTime:
        """Retrieve values of a single custom metric over a time period.

        Parameters
        ----------
        custom_metric_id: str
            The ID of the custom metric.
        deployment_id: str
            The ID of the deployment.
        start: datetime or str
            Start of the time period.
        end: datetime or str
            End of the time period.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        bucket_size: Optional[str]
            Time duration of a bucket, in ISO 8601 time duration format.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_over_time: CustomMetricValuesOverTime
            The queried custom metric values over time information.
        """
        path = cls._path.format(deployment_id, custom_metric_id)
        params = cls._build_query_params(
            start=start,
            end=end,
            model_id=model_id,
            model_package_id=model_package_id,
            bucket_size=bucket_size,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        custom_metric_over_time = cls.from_data(case_converted)
        custom_metric_over_time.deployment_id = deployment_id
        return custom_metric_over_time

    @property
    def bucket_values(self) -> Dict[datetime, int]:
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: Dict
        """
        if self.buckets:
            return {
                bucket["period"]["start"]: bucket["value"]
                for bucket in self.buckets
                if bucket.get("period")
            }
        return {}

    @property
    def bucket_sample_sizes(self) -> Dict[datetime, int]:
        """The sample size for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_sample_sizes: Dict
        """
        if self.buckets:
            return {
                bucket["period"]["start"]: bucket["sample_size"]
                for bucket in self.buckets
                if bucket.get("period")
            }
        return {}

    def get_buckets_as_dataframe(self) -> pd.DataFrame:
        """Retrieves all custom metrics buckets in a pandas DataFrame.

        Returns
        -------
        buckets: pd.DataFrame
        """
        if self.buckets:
            rows = []
            for bucket in self.buckets:
                rows.append(
                    {
                        "start": bucket["period"]["start"],
                        "end": bucket["period"]["end"],
                        "value": bucket["value"],
                        "sample_size": bucket["sample_size"],
                    }
                )
            return pd.DataFrame(rows)
        return pd.DataFrame()


class CustomMetricSummary(APIObject, MonitoringDataQueryBuilderMixin):
    """The summary of a custom metric.

    .. versionadded:: v3.4

    Attributes
    ----------
    period: Period
        A time period defined by a start and end tie
    metric: Dict
        The summary of the custom metric.
    """

    _path = "deployments/{}/customMetrics/{}/summary/"
    _period = t.Dict(
        {
            t.Key("start"): t.String >> dateutil.parser.parse,
            t.Key("end"): t.String >> dateutil.parser.parse,
        }
    ).allow_extra("*")
    _metric = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("value"): t.Or(t.Float, t.Null),
            t.Key("sample_count"): t.Or(t.Int, t.Null),
            t.Key("baseline_value"): t.Or(t.Float, t.Null),
            t.Key("percent_change"): t.Or(t.Float, t.Null),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metric"): _metric,
        }
    ).allow_extra("*")

    def __init__(
        self,
        period: Period,
        metric: Dict[str, Any],
        deployment_id: Optional[str] = None,
    ):
        self.period = period
        self.metric = metric
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "CustomMetricSummary({} - {}: {})".format(
            self.period.get("start"), self.period.get("end"), self.metric
        )

    @classmethod
    def get(
        cls,
        deployment_id: str,
        custom_metric_id: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricSummary:
        """Retrieve the summary of a custom metric over a time period.

        Parameters
        ----------
        custom_metric_id: str
            The ID of the custom metric.
        deployment_id: str
            The ID of the deployment.
        start: datetime or str
            Start of the time period.
        end: datetime or str
            End of the time period.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_summary: CustomMetricSummary
            The summary of the custom metric.
        """
        path = cls._path.format(deployment_id, custom_metric_id)
        params = cls._build_query_params(
            start=start,
            end=end,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        custom_metric_summary = cls.from_data(case_converted)
        custom_metric_summary.deployment_id = deployment_id
        return custom_metric_summary


class CustomMetricValuesOverBatch(APIObject, MonitoringDataQueryBuilderMixin):
    """Custom metric over batch information.

    .. versionadded:: v3.4

    Attributes
    ----------
    buckets: List[BatchBucket]
        A list of buckets with custom metric values aggregated over batches.
    metric: Dict
        A custom metric definition.
    deployment_id: str
        The ID of the deployment.
    segment_attribute: str
        The name of the segment on which segment analysis is being performed.
    segment_value: str
        The value of the segment_attribute to segment on.
    """

    _path = "deployments/{}/customMetrics/{}/valuesOverBatch/"
    _batch = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("created_at"): t.String >> dateutil.parser.parse,
            t.Key("last_prediction_timestamp", optional=True): t.String >> dateutil.parser.parse,
        }
    ).allow_extra("*")
    _bucket = t.Dict(
        {
            t.Key("batch"): t.Or(_batch),
            t.Key("value"): t.Or(t.Float, t.Null),
            t.Key("sample_size"): t.Or(t.Int, t.Null),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("buckets"): t.List(_bucket),
            t.Key("metric"): t.Dict().allow_extra("*"),
            t.Key("segment_attribute", optional=True): t.String(),
            t.Key("segment_value", optional=True): t.String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        buckets: Optional[List[BatchBucket]] = None,
        metric: Optional[Dict[str, Any]] = None,
        deployment_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ):
        self.buckets = buckets if buckets is not None else []
        self.metric = metric
        self.deployment_id = deployment_id
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        first_batch = self.buckets[0]["batch"].get("id") if self.buckets else None
        last_batch = self.buckets[-1]["batch"].get("id") if self.buckets else None
        return "CustomMetricValuesOverBatch({} - {})".format(
            first_batch,
            last_batch,
        )

    @classmethod
    def get(
        cls,
        deployment_id: str,
        custom_metric_id: str,
        batch_ids: Optional[List[str]] = None,
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricValuesOverBatch:
        """Retrieve values of a single custom metric over batches.

        Parameters
        ----------
        custom_metric_id: str
            The ID of the custom metric.
        deployment_id: str
            The ID of the deployment.
        batch_ids : Optional[List[str]]
            Specify a list of batch IDs to pull the data for.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_over_batch: CustomMetricValuesOverBatch
            The queried custom metric values over batch information.
        """
        path = cls._path.format(deployment_id, custom_metric_id)
        params = cls._build_query_params(
            batch_id=batch_ids,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        custom_metric_over_batch = cls.from_data(case_converted)
        custom_metric_over_batch.deployment_id = deployment_id
        return custom_metric_over_batch

    @property
    def bucket_values(self) -> Dict[str, int]:
        """The metric value for all batch buckets, keyed by batch ID

        Returns
        -------
        bucket_values: Dict
        """
        if self.buckets:
            return {bucket["batch"]["id"]: bucket["value"] for bucket in self.buckets}
        return {}

    @property
    def bucket_sample_sizes(self) -> Dict[str, int]:
        """The sample size for all batch buckets, keyed by batch ID.

        Returns
        -------
        bucket_sample_sizes: Dict
        """
        if self.buckets:
            return {bucket["batch"]["id"]: bucket["sample_size"] for bucket in self.buckets}
        return {}

    def get_buckets_as_dataframe(self) -> pd.DataFrame:
        """Retrieves all custom metrics buckets in a pandas DataFrame.

        Returns
        -------
        buckets: pd.DataFrame
        """
        if self.buckets:
            rows = []
            for bucket in self.buckets:
                rows.append(
                    {
                        "batch_id": bucket["batch"]["id"],
                        "batch_name": bucket["batch"]["name"],
                        "value": bucket["value"],
                        "sample_size": bucket["sample_size"],
                    }
                )
            return pd.DataFrame(rows)
        return pd.DataFrame()


class CustomMetricBatchSummary(APIObject, MonitoringDataQueryBuilderMixin):
    """The batch summary of a custom metric.

    .. versionadded:: v3.4

    Attributes
    ----------
    metric: Dict
        The summary of the batch custom metric.
    """

    _path = "deployments/{}/customMetrics/{}/batchSummary/"
    _metric = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("value"): t.Or(t.Float, t.Null),
            t.Key("sample_count"): t.Or(t.Int, t.Null),
            t.Key("baseline_value"): t.Or(t.Float, t.Null),
            t.Key("percent_change"): t.Or(t.Float, t.Null),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("metric"): _metric,
        }
    ).allow_extra("*")

    def __init__(
        self,
        metric: Dict[str, Any],
        deployment_id: Optional[str] = None,
    ):
        self.metric = metric
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "CustomMetricBatchSummary({})".format(self.metric)

    @classmethod
    def get(
        cls,
        deployment_id: str,
        custom_metric_id: str,
        batch_ids: Optional[List[str]] = None,
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> CustomMetricBatchSummary:
        """Retrieve the summary of a custom metric over a batch.

        Parameters
        ----------
        custom_metric_id: str
            The ID of the custom metric.
        deployment_id: str
            The ID of the deployment.
        batch_ids : Optional[List[str]]
            Specify a list of batch IDs to pull the data for.
        model_id: Optional[str]
            The ID of the model.
        model_package_id: Optional[str]
            The ID of the model package.
        segment_attribute: Optional[str]
            The name of the segment on which segment analysis is being performed.
        segment_value: Optional[str]
            The value of the segment_attribute to segment on.

        Returns
        -------
        custom_metric_summary: CustomMetricBatchSummary
            The batch summary of the custom metric.
        """
        path = cls._path.format(deployment_id, custom_metric_id)
        params = cls._build_query_params(
            batch_id=batch_ids,
            model_id=model_id,
            model_package_id=model_package_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        custom_metric_summary = cls.from_data(case_converted)
        custom_metric_summary.deployment_id = deployment_id
        return custom_metric_summary


class HostedCustomMetricTemplate(APIObject):
    """
    Template for hosted custom metric.
    """

    _path = "customMetricsTemplates/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("description"): t.String(),
            t.Key("custom_metric_metadata"): t.Dict().allow_extra("*"),
            t.Key("default_environment"): t.Dict().allow_extra("*"),
            t.Key("items"): t.List(t.Dict().allow_extra("*")),
            t.Key("template_metric_type"): t.String(),
        }
    )

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        custom_metric_metadata: Dict[str, Any],
        default_environment: Dict[str, Any],
        items: List[Dict[str, Any]],
        template_metric_type: str,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.custom_metric_metadata = custom_metric_metadata
        self.default_environment = default_environment
        self.items = items
        self.template_metric_type = template_metric_type

    @classmethod
    def list(
        cls,
        search: Optional[str] = None,
        order_by: Optional[ListHostedCustomMetricTemplatesSortQueryParams] = None,
        metric_type: Optional[HostedCustomMetricsTemplateMetricTypeQueryParams] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[HostedCustomMetricTemplate]:
        """List all hosted custom metric templates.

        Parameters
        ----------
        search: Optional[str]
            Search string.
        order_by: Optional[ListHostedCustomMetricTemplatesSortQueryParams]
            Ordering field.
        metric_type: Optional[HostedCustomMetricsTemplateMetricTypeQueryParams]
            Type of the metric.
        offset: Optional[int]
            Offset for pagination.
        limit: Optional[int]
            Limit for pagination.

        Returns
        -------
        templates: List[HostedCustomMetricTemplate]
        """
        params: Dict[str, Any] = {}
        if search:
            params["search"] = search
        if order_by:
            params["orderBy"] = order_by
        if metric_type:
            params["templateMetricType"] = metric_type
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit
        response = cls._client.get(cls._path, params=params if params else None)
        return [cls.from_server_data(d) for d in response.json()["data"]]

    @classmethod
    def get(cls, template_id: str) -> HostedCustomMetricTemplate:
        """Get a hosted custom metric template by ID.

        Parameters
        ----------
        template_id: str
            ID of the template.

        Returns
        -------
        template : HostedCustomMetricTemplate
        """
        response = cls._client.get(f"{cls._path}{template_id}/")
        return cls.from_server_data(response.json())


class MetricTimestampSpoofing:
    """
    Custom metric timestamp spoofing. Occurs when reading values from a file, like a dataset.
    By default, replicates pd.to_datetime formatting behavior.
    """

    def __init__(self, column_name: str, time_format: Optional[str] = None):
        self.column_name = column_name
        self.time_format = time_format


class BatchField:
    """
    A custom metric batch ID source for when reading values from a columnar dataset like a file.
    """

    def __init__(self, column_name: str):
        self.column_name = column_name


class ValueField:
    """
    A custom metric value source for when reading values from a columnar dataset like a file.
    """

    def __init__(self, column_name: str):
        self.column_name = column_name


class SampleCountField:
    """
    A weight column used with columnar datasets if pre-aggregated metric values are provided.
    """

    def __init__(self, column_name: str):
        self.column_name = column_name


class MetricBaselineValue:
    """
    The baseline values for a custom metric.
    """

    def __init__(self, value: float):
        self.value = value


class DeploymentDetails:
    """
    Information about a hosted custom metric deployment.
    """

    def __init__(
        self,
        id: str,
        name: str,
        creator_first_name: Optional[str] = None,
        creator_last_name: Optional[str] = None,
        creator_username: Optional[str] = None,
        creator_gravatar_hash: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.creator_first_name = creator_first_name
        self.creator_last_name = creator_last_name
        self.creator_username = creator_username
        self.creator_gravatar_hash = creator_gravatar_hash
        self.created_at = created_at


class HostedCustomMetric(APIObject):
    """
    Hosted custom metric.
    """

    _schedule = t.Dict(
        {
            t.Key("hour"): t.List(t.Or(t.String(), t.Int())),
            t.Key("minute"): t.List(t.Or(t.String(), t.Int())),
            t.Key("day_of_week"): t.List(t.Or(t.String(), t.Int())),
            t.Key("day_of_month"): t.List(t.Or(t.String(), t.Int())),
            t.Key("month"): t.List(t.Or(t.String(), t.Int())),
        }
    ).allow_extra("*")
    _parameter_overrides = t.Dict(
        {
            t.Key("field_name"): t.String(),
            t.Key("value", optional=True): t.Or(t.String(), t.Int(), t.Float(), t.Bool(), t.Null()),
            t.Key("type"): t.String(),
        }
    ).allow_extra("*")
    _baseline_values = t.Dict(
        {
            t.Key("value"): t.Float,
        }
    ).allow_extra("*")
    _timestamp = t.Dict(
        {
            t.Key("column_name"): t.String(),
            t.Key("time_format", optional=True): t.String(),
        }
    ).allow_extra("*")
    _value = t.Dict(
        {
            t.Key("column_name"): t.String(),
        }
    ).allow_extra("*")
    _sample_count = t.Dict(
        {
            t.Key("column_name"): t.String(),
        }
    ).allow_extra("*")
    _batch = t.Dict(
        {
            t.Key("column_name"): t.String(),
        }
    ).allow_extra("*")
    _deployment = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("created_at", optional=True): t.String(),
            t.Key("creator_first_name", optional=True): t.String(),
            t.Key("creator_last_name", optional=True): t.String(),
            t.Key("creator_username", optional=True): t.String(),
            t.Key("creator_gravatar_hash", optional=True): t.String(),
        }
    ).allow_extra("*")

    _user = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("username", optional=True): t.String(),
            t.Key("first_name", optional=True): t.String(),
            t.Key("last_name", optional=True): t.String(),
            t.Key("gravatar_hash", optional=True): t.String(),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("deployment"): _deployment,
            t.Key("schedule", optional=True): t.Or(_schedule, t.Null()),
            t.Key("units"): t.String(),
            t.Key("type"): t.String(),
            t.Key("is_model_specific"): t.Bool(),
            t.Key("directionality"): t.String(),
            t.Key("time_step"): t.String(),
            t.Key("created_at"): t.String(),
            t.Key("created_by"): t.Dict().allow_extra("*"),
            t.Key("name"): t.String(),
            t.Key(
                "description",
                optional=True,
            ): t.Or(t.String(allow_blank=True), t.Null()),
            t.Key("baseline_values", optional=True): t.Or(
                t.List(t.Dict().allow_extra("*")), t.Null()
            ),
            t.Key("timestamp", optional=True): t.Or(_timestamp, t.Null()),
            t.Key("value", optional=True): t.Or(_value, t.Null()),
            t.Key("sample_count", optional=True): t.Or(_sample_count, t.Null()),
            t.Key("batch", optional=True): t.Or(_batch, t.Null()),
            t.Key("parameter_overrides", optional=True): t.Or(
                t.List(_parameter_overrides), t.Null()
            ),
            t.Key("custom_job_id"): t.String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        deployment: Dict[str, Any],
        units: str,
        type: str,
        is_model_specific: bool,
        directionality: str,
        time_step: str,
        created_at: str,
        created_by: Dict[str, Any],
        name: str,
        custom_job_id: str,
        description: Optional[str] = None,
        schedule: Optional[Schedule] = None,
        baseline_values: Optional[List[Dict[str, Any]]] = None,
        timestamp: Optional[Dict[str, Any]] = None,
        value: Optional[Dict[str, Any]] = None,
        sample_count: Optional[Dict[str, Any]] = None,
        batch: Optional[Dict[str, Any]] = None,
        parameter_overrides: Optional[List[Dict[str, Any]]] = None,
    ):
        """

        Parameters
        ----------
        id: str
            ID of the hosted custom metric.
        deployment: Dict[str, Any]
            Deployment details.
        units: str
            Units of the metric.
        type: str
            Type of the metric.
        is_model_specific: bool
            Whether the metric is model specific.
        directionality: str
            Directionality of the metric.
        time_step: str
            Time step of the metric.
        created_at: str
            Creation time of the metric.
        created_by: Dict[str, Any]
            Creator details.
        name: str
            Name of the metric.
        custom_job_id: str
            ID of the custom job connected to this hosted custom metric
        description: Optional[str]
            Description of the metric.
        schedule: Optional[Dict[str, Any]]
            Schedule details.
        baseline_values: Optional[List[Dict[str, Any]]
            Baseline values.
        timestamp: Optional[Dict[str, Any]]
            Timestamp details.
        value: Optional[Dict[str, Any]]
            Value details.
        sample_count: Optional[Dict[str, Any]]
            Sample count details.
        batch: Optional[Dict[str, Any]]
            Batch details.
        parameter_overrides: Optional[List[Dict[str, Any]]
            Parameter overrides.
        """
        self.id = id
        self.deployment = DeploymentDetails(
            id=deployment["id"],
            name=deployment["name"],
            creator_first_name=deployment.get("creator_first_name"),
            creator_last_name=deployment.get("creator_last_name"),
            creator_username=deployment.get("creator_username"),
            creator_gravatar_hash=deployment.get("creator_gravatar_hash"),
            created_at=deployment.get("created_at"),
        )
        self.schedule = (
            Schedule(
                hour=schedule["hour"],
                minute=schedule["minute"],
                day_of_week=schedule["day_of_week"],
                day_of_month=schedule["day_of_month"],
                month=schedule["month"],
            )
            if schedule
            else None
        )
        self.units = units
        self.type = type
        self.is_model_specific = is_model_specific
        self.directionality = directionality
        self.time_step = time_step
        self.created_at = created_at
        self.created_by = created_by
        self.name = name
        self.custom_job_id = custom_job_id
        self.description = description
        self.baseline_values = (
            [MetricBaselineValue(**value) for value in baseline_values] if baseline_values else None
        )
        self.timestamp = (
            MetricTimestampSpoofing(
                column_name=timestamp["column_name"], time_format=timestamp.get("time_format")
            )
            if timestamp
            else None
        )
        self.value = ValueField(column_name=value["column_name"]) if value else None
        self.sample_count = (
            SampleCountField(column_name=sample_count["column_name"]) if sample_count else None
        )
        self.batch = BatchField(column_name=batch["column_name"]) if batch else None
        parameter_overrides = self._impute_null_runtime_parameter_values(parameter_overrides)
        self.parameter_overrides = (
            [
                RuntimeParameterValue(
                    field_name=param["field_name"], value=param["value"], type=param["type"]
                )
                for param in parameter_overrides
            ]
            if parameter_overrides
            else None
        )

    def _impute_null_runtime_parameter_values(
        self, values: Optional[List[Dict[str, Any]]]
    ) -> Optional[List[Dict[str, Any]]]:
        if not values:
            return values
        for param in values:
            if "value" not in values:
                param["value"] = None
        return values

    @classmethod
    def list(
        cls,
        job_id: str,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[HostedCustomMetric]:
        """List all hosted custom metrics for a job.

        Parameters
        ----------
        job_id: str
            ID of the job.

        Returns
        -------
        metrics: List[HostedCustomMetric]
        """
        params: Dict[str, Any] = {}
        if skip:
            params["skip"] = skip
        if limit:
            params["limit"] = limit
        response = cls._client.get(f"customJobs/{job_id}/customMetrics/", params=params)
        return [cls.from_server_data(d) for d in response.json()["data"]]

    @classmethod
    def create_from_template(
        cls,
        template_id: str,
        deployment_id: str,
        job_name: str,
        custom_metric_name: str,
        job_description: Optional[str] = None,
        custom_metric_description: Optional[str] = None,
        sidecar_deployment_id: Optional[str] = None,
        baseline_value: Optional[float] = None,
        timestamp: Optional[MetricTimestampSpoofing] = None,
        value: Optional[ValueField] = None,
        sample_count: Optional[SampleCountField] = None,
        batch: Optional[BatchField] = None,
        schedule: Optional[Schedule] = None,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> HostedCustomMetric:
        """Create a hosted custom metric from a template.
        A shortcut for 2 calls:
        Job.from_custom_metric_template(template_id)
        HostedCustomMetrics.create_from_custom_job()

        Parameters
        ----------
        template_id: str
            ID of the template.
        deployment_id: str
            ID of the deployment.
        job_name: str
            Name of the job.
        custom_metric_name: str
            Name of the metric.
        job_description: Optional[str]
            Description of the job.
        custom_metric_description: Optional[str]
            Description of the metric.
        sidecar_deployment_id: Optional[str]
            ID of the sidecar deployment.
        baseline_value: Optional[float]
            Baseline value.
        timestamp: Optional[MetricTimestampSpoofing]
            Timestamp details.
        value: Optional[ValueField]
            Value details.
        sample_count: Optional[SampleCountField]
            Sample count details.
        batch: Optional[BatchField]
            Batch details.
        schedule: Optional[Schedule]
            Schedule details.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            Parameter overrides.

        Returns
        -------
        metric: HostedCustomMetric
        """
        job = Job.create_from_custom_metric_gallery_template(
            template_id=template_id,
            name=job_name,
            description=job_description,
            sidecar_deployment_id=sidecar_deployment_id,
        )
        hosted_custom_metric = cls.create_from_custom_job(
            custom_job_id=job.id,
            deployment_id=deployment_id,
            name=custom_metric_name,
            description=custom_metric_description,
            baseline_value=baseline_value,
            timestamp=timestamp,
            value=value,
            sample_count=sample_count,
            batch=batch,
            schedule=schedule,
            parameter_overrides=parameter_overrides,
        )
        return hosted_custom_metric

    @classmethod
    def create_from_custom_job(
        cls,
        custom_job_id: str,
        deployment_id: str,
        name: str,
        description: Optional[str] = None,
        baseline_value: Optional[float] = None,
        timestamp: Optional[MetricTimestampSpoofing] = None,
        value: Optional[ValueField] = None,
        sample_count: Optional[SampleCountField] = None,
        batch: Optional[BatchField] = None,
        schedule: Optional[Schedule] = None,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> HostedCustomMetric:
        """Create a hosted custom metric from existing custom job.

        Parameters
        ----------
        custom_job_id: str
            ID of the custom job.
        deployment_id: str
            ID of the deployment.
        name: str
            Name of the metric.
        description: Optional[str]
            Description of the metric.
        baseline_value: Optional[float]
            Baseline value.
        timestamp: Optional[MetricTimestampSpoofing]
            Timestamp details.
        value: Optional[ValueField]
            Value details.
        sample_count: Optional[SampleCountField]
            Sample count details.
        batch: Optional[BatchField]
            Batch details.
        schedule: Optional[Schedule]
            Schedule details.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            Parameter overrides.

        Returns
        -------
        metric: HostedCustomMetric

        """
        url = f"deployments/{deployment_id}/customMetrics/fromCustomJob/"
        payload: Dict[str, Any] = {
            "customJobId": custom_job_id,
            "name": name,
        }
        if description:
            payload["description"] = description
        if timestamp:
            payload["timestamp"] = {"columnName": timestamp.column_name}
            if timestamp.time_format:
                payload["timestamp"]["timeFormat"] = timestamp.time_format
        if value:
            payload["value"] = {"columnName": value.column_name}
        if baseline_value:
            payload["baselineValues"] = [{"value": baseline_value}]
        if sample_count:
            payload["sampleCount"] = {"columnName": sample_count.column_name}
        if batch:
            payload["batch"] = {"columnName": batch.column_name}
        if schedule:
            payload["schedule"] = to_api(schedule)
        if parameter_overrides:
            payload["parameterOverrides"] = [
                {"fieldName": param.field_name, "value": param.value, "type": param.type}
                for param in parameter_overrides
            ]
        result = cls._client.post(url, json=payload)
        return cls.from_server_data(result.json())

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        units: Optional[str] = None,
        directionality: Optional[str] = None,
        aggregation_type: Optional[CustomMetricAggregationType] = None,
        baseline_value: Optional[List[float]] = None,
        timestamp: Optional[MetricTimestampSpoofing] = None,
        value: Optional[ValueField] = None,
        sample_count: Optional[SampleCountField] = None,
        batch: Optional[BatchField] = None,
        schedule: Optional[Schedule] = None,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> HostedCustomMetric:
        """Update the hosted custom metric.

        Parameters
        ----------
        name: Optional[str]
            Name of the metric.
        description: Optional[str]
            Description of the metric.
        units: Optional[str]
            Units of the metric.
        directionality: Optional[str]
            Directionality of the metric.
        aggregation_type: Optional[CustomMetricAggregationType]
            Aggregation type of the metric.
        baseline_value: Optional[float]
            Baseline values.
        timestamp: Optional[MetricTimestampSpoofing]
            Timestamp details.
        value: Optional[ValueField]
            Value details.
        sample_count: Optional[SampleCountField]
            Sample count details.
        batch: Optional[BatchField]
            Batch details.
        schedule: Optional[Schedule]
            Schedule details.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            Parameter overrides.

        Returns
        -------
        metric: HostedCustomMetric
        """
        url = f"customJobs/{self.custom_job_id}/customMetrics/{self.id}/"
        payload: Dict[str, Any] = {}
        if name:
            payload["name"] = name
        if description:
            payload["description"] = description
        if units:
            payload["units"] = units
        if directionality:
            payload["directionality"] = directionality
        if aggregation_type:
            payload["aggregationType"] = aggregation_type
        if baseline_value:
            payload["baselineValues"] = [{"value": baseline_value}]
        if timestamp:
            payload["timestamp"] = {"columnName": timestamp.column_name}
            if timestamp.time_format:
                payload["timestamp"]["timeFormat"] = timestamp.time_format
        if value:
            payload["value"] = {"columnName": value.column_name}
        if sample_count:
            payload["sampleCount"] = {"columnName": sample_count.column_name}
        if batch:
            payload["batch"] = {"columnName": batch.column_name}
        if schedule:
            payload["schedule"] = schedule
        if parameter_overrides:
            payload["parameter_overrides"] = [
                {"fieldName": param.field_name, "value": param.value, "type": param.type}
                for param in parameter_overrides
            ]

        response = self._client.patch(url, json=payload)
        return self.from_server_data(response.json())

    def delete(self) -> None:
        """Delete the hosted custom metric."""
        url = f"customJobs/{self.custom_job_id}/customMetrics/{self.id}/"
        self._client.delete(url)


class HostedCustomMetricBlueprint(APIObject):
    """
    Hosted custom metric blueprints provide an option to share custom metric settings between multiple
    custom metrics sharing the same custom jobs. When a custom job of a hosted custom metric type is connected to the
    deployment, all the custom metric parameters from the blueprint are automatically copied.
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("directionality"): t.String(),
            t.Key("units"): t.String(),
            t.Key("type"): t.String(),
            t.Key("time_step"): t.String(),
            t.Key("is_model_specific"): t.Bool(),
            t.Key("custom_job_id"): t.String(),
            t.Key("created_at"): t.String(),
            t.Key("updated_at"): t.String(),
            t.Key("created_by"): t.String(),
            t.Key("updated_by"): t.String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        directionality: str,
        units: str,
        type: str,
        time_step: str,
        is_model_specific: bool,
        custom_job_id: str,
        created_at: str,
        updated_at: str,
        created_by: str,
        updated_by: str,
    ):
        """
        Parameters
        ----------
        id: str
            ID of the hosted custom metric blueprint.
        directionality: str
            Directionality of the metric.
        units: str
            Units of the metric.
        type: str
            Type of the metric.
        time_step: str
            Time step of the metric.
        is_model_specific: bool
            Sets whether the metric is model specific.
        custom_job_id: str
            ID of the custom job connected to this hosted custom metric
        created_at: str
            Creation time of the metric.
        updated_at: str
            Update time of the metric.
        created_by: str
            Creator of the metric.
        updated_by: str
            Updater of the metric.

        """
        self.id = id
        self.units = units
        self.type = type
        self.is_model_specific = is_model_specific
        self.directionality = directionality
        self.time_step = time_step
        self.custom_job_id = custom_job_id
        self.created_at = created_at
        self.updated_at = updated_at
        self.created_by = created_by
        self.updated_by = updated_by

    @classmethod
    def get(cls, custom_job_id: str) -> HostedCustomMetricBlueprint:
        """Get a hosted custom metric blueprint.

        Parameters
        ----------
        custom_job_id: str
            ID of the custom job.

        Returns
        -------
        blueprint: HostedCustomMetricBlueprint
        """
        url = f"customJobs/{custom_job_id}/hostedCustomMetricTemplate/"
        response = cls._client.get(url)
        return cls.from_server_data(response.json())

    @classmethod
    def create(
        cls,
        custom_job_id: str,
        directionality: str,
        units: str,
        type: str,
        time_step: str,
        is_model_specific: bool,
    ) -> HostedCustomMetricBlueprint:
        """Create a hosted custom metric blueprint.

        Parameters
        ----------
        custom_job_id: str
            ID of the custom job.
        directionality: str
            Directionality of the metric.
        units: str
            Units of the metric.
        type: str
            Type of the metric.
        time_step: str
            Time step of the metric.
        is_model_specific: bool
            Whether the metric is model specific.

        Returns
        -------
        blueprint: HostedCustomMetricBlueprint
        """
        url = f"customJobs/{custom_job_id}/hostedCustomMetricTemplate/"
        payload = {
            "directionality": directionality,
            "units": units,
            "type": type,
            "timeStep": time_step,
            "isModelSpecific": is_model_specific,
        }
        response = cls._client.post(url, json=payload)
        return cls.from_server_data(response.json())

    def update(
        self,
        directionality: Optional[str] = None,
        units: Optional[str] = None,
        type: Optional[str] = None,
        time_step: Optional[str] = None,
        is_model_specific: Optional[bool] = None,
    ) -> HostedCustomMetricBlueprint:
        """Update a hosted custom metric blueprint.

        Parameters
        ----------
        directionality: Optional[str]
            Directionality of the metric.
        units: Optional[str]
            Units of the metric.
        type: Optional[str]
            Type of the metric.
        time_step: Optional[str]
            Time step of the metric.
        is_model_specific: Optional[bool]
            Determines whether the metric is model specific.

        Returns
        -------
        updated_blueprint: HostedCustomMetricBlueprint

        """
        url = f"customJobs/{self.custom_job_id}/hostedCustomMetricTemplate/"
        payload: Dict[str, Any] = {}
        if directionality:
            payload["directionality"] = directionality
        if units:
            payload["units"] = units
        if type:
            payload["type"] = type
        if time_step:
            payload["timeStep"] = time_step
        if is_model_specific:
            payload["isModelSpecific"] = is_model_specific
        response = self._client.patch(url, json=payload)
        return self.from_server_data(response.json())



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate

playground_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("description"): t.String(allow_blank=True),
        t.Key("use_case_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("last_update_date"): t.String,
        t.Key("last_update_user_id"): t.String,
        t.Key("saved_llm_blueprints_count"): t.Int,
        t.Key("llm_blueprints_count"): t.Int,
        t.Key("user_name"): t.String(allow_blank=True),
    }
).ignore_extra("*")


class Playground(APIObject):
    """
    Metadata for a DataRobot GenAI playground.

    Attributes
    ----------
    id : str
        Playground ID.
    name : str
        Playground name.
    description : str
        Description of the playground.
    use_case_id : str
        Linked use case ID.
    creation_date : str
        The date when the playground was created.
    creation_user_id : str
        ID of the creating user.
    last_update_date : str
        Date when the playground was most recently updated.
    last_update_user_id : str
        ID of the user who most recently updated the playground.
    saved_llm_blueprints_count : int
        Number of saved LLM blueprints in the playground.
    llm_blueprints_count : int
        Number of LLM blueprints in the playground.
    user_name : str
        The name of the user who created the playground.
    """

    _path = "api/v2/genai/playgrounds"

    _converter = playground_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        use_case_id: str,
        creation_date: str,
        creation_user_id: str,
        last_update_date: str,
        last_update_user_id: str,
        saved_llm_blueprints_count: int,
        llm_blueprints_count: int,
        user_name: str,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.use_case_id = use_case_id
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.last_update_date = last_update_date
        self.last_update_user_id = last_update_user_id
        self.saved_llm_blueprints_count = saved_llm_blueprints_count
        self.llm_blueprints_count = llm_blueprints_count
        self.user_name = user_name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name})"

    @classmethod
    def create(
        cls,
        name: str,
        description: str = "",
        use_case: Optional[Union[UseCase, str]] = None,
    ) -> Playground:
        """
        Create a new playground.

        Parameters
        ----------
        name : str
            Playground name.
        description : str, optional
            Description of the playground, by default "".
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created playground.

        Returns
        -------
        playground : Playground
            The created playground.
        """
        payload = {
            "name": name,
            "description": description,
            "use_case_id": get_use_case_id(use_case, is_required=True),
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, playground_id: str) -> Playground:
        """
        Retrieve a single playground.

        Parameters
        ----------
        playground_id : str
            The ID of the playground you want to retrieve.

        Returns
        -------
        playground : Playground
            The requested playground.
        """
        url = f"{cls._client.domain}/{cls._path}/{playground_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        use_case: Optional[UseCaseLike] = None,
        search: Optional[str] = None,
        sort: Optional[str] = None,
    ) -> List[Playground]:
        """
        List all playgrounds available to the user. If the use_case is specified or can be
        inferred from the Context then the results are restricted to the playgrounds
        associated with the UseCase.

        Parameters
        ----------
        use_case : Optional[UseCaseLike], optional
            The returned playgrounds are filtered to those associated with a specific Use Case
            or Cases if specified or can be inferred from the Context.
            Accepts either the entity or the ID.
        search : str, optional
            String for filtering playgrounds.
            Playgrounds that contain the string in name will be returned.
            If not specified, all playgrounds will be returned.
        sort : str, optional
            Property to sort playgrounds by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-creationDate'.
            Currently supported options are listed in ListPlaygroundsSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            playgrounds being returned in order of creation time descending.

        Returns
        -------
        playgrounds : list[Playground]
            A list of playgrounds available to the user.
        """
        params = {
            "search": search,
            "sort": sort,
        }
        params = resolve_use_cases(use_cases=use_case, params=params, use_case_key="use_case_id")
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def update(self, name: Optional[str] = None, description: Optional[str] = None) -> Playground:
        """
        Update the playground.

        Parameters
        ----------
        name : str
            The new name for the playground.
        description: str
            The new description for the playground.

        Returns
        -------
        playground : Playground
            The updated playground.
        """
        payload = {"name": name, "description": description}
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())

    def delete(self) -> None:
        """
        Delete the playground.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, cast, Dict, List, Optional, Type, TypeVar

import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import (
    CUSTOM_MODEL_TARGET_TYPE,
    DEFAULT_MAX_WAIT,
    NETWORK_EGRESS_POLICY,
    TARGET_TYPE,
)
from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model_version import CustomModelVersion
from datarobot.utils import deprecated
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

T_CustomModelBase = TypeVar("T_CustomModelBase", bound="_CustomModelBase")


class _CustomModelBase(APIObject):  # pylint: disable=missing-class-docstring
    _path = "customModels/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("description"): String(allow_blank=True),
            t.Key("supports_binary_classification", optional=True, default=False): t.Bool(),
            t.Key("supports_regression", optional=True, default=False): t.Bool(),
            t.Key("supports_textgeneration", optional=True, default=False): t.Bool(),
            t.Key("latest_version", optional=True, default=None): t.Or(
                CustomModelVersion.schema, t.Null()
            ),
            t.Key("deployments_count", optional=True, default=None): Int(),
            t.Key("created_by"): String(),
            t.Key("updated") >> "updated_at": String(),
            t.Key("created") >> "created_at": String(),
            t.Key("target_type", optional=True, default=None): String(),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id: str,
        name: str,
        description: str,
        supports_binary_classification: bool,
        supports_regression: bool,
        supports_textgeneration: bool,
        latest_version: CustomModelVersion,
        deployments_count: int,
        created_by: str,
        updated_at: str,
        created_at: str,
        target_type: str,
    ) -> None:
        self.id = id
        self.name = name
        self.description = description

        self.target_type = target_type
        if supports_binary_classification + supports_regression + supports_textgeneration > 1:
            raise ValueError("Model should support only 1 target type")

        if not target_type:
            if supports_binary_classification:
                self.target_type = CUSTOM_MODEL_TARGET_TYPE.BINARY
            elif supports_regression:
                self.target_type = CUSTOM_MODEL_TARGET_TYPE.REGRESSION
            elif supports_textgeneration:
                self.target_type = CUSTOM_MODEL_TARGET_TYPE.TEXT_GENERATION
            else:
                raise ValueError("Target type must be provided")
        else:
            if target_type != CUSTOM_MODEL_TARGET_TYPE.BINARY and supports_binary_classification:
                raise ValueError(
                    "Cannot specify both target_type {} and "
                    "supports_binary_classification.".format(target_type)
                )
            elif target_type != CUSTOM_MODEL_TARGET_TYPE.REGRESSION and supports_regression:
                raise ValueError(
                    "Cannot specify both target_type {} and "
                    "supports_regression.".format(target_type)
                )
            elif (
                target_type != CUSTOM_MODEL_TARGET_TYPE.TEXT_GENERATION and supports_textgeneration
            ):
                raise ValueError(
                    "Cannot specify both target_type {} and "
                    "supports_text_generation.".format(target_type)
                )

        self.latest_version = CustomModelVersion(**latest_version) if latest_version else None  # type: ignore[arg-type]
        self.deployments_count = deployments_count
        self.created_by = created_by
        self.updated_at = updated_at
        self.created_at = created_at

    @classmethod
    def _check_model_type(cls, data: Dict[str, Any]) -> bool:
        return cast(bool, data["customModelType"] == cls._model_type)  # type: ignore[attr-defined]

    @classmethod
    def list(
        cls: Type[T_CustomModelBase],
        is_deployed: Optional[bool] = None,
        order_by: Optional[str] = None,
        search_for: Optional[str] = None,
    ) -> List[T_CustomModelBase]:
        """List instances of _CustomModelBase.

        .. versionadded:: v2.21

        Parameters
        ----------
        is_deployed: bool, optional
            Flag for filtering custom inference models.
            If set to `True`, only deployed custom inference models are returned.
            If set to `False`, only not deployed custom inference models are returned.
        search_for: str, optional
            String for filtering custom inference models - only custom.
            inference models that contain the string in name or description will
            be returned.
            If not specified, all custom models will be returned.
        order_by: str, optional
            Property to sort custom inference models by.
            Supported properties are "created" and "updated".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None which will result in
            custom models being returned in order of creation time descending

        Returns
        -------
        List[T_CustomModelBase]
            A list of instances of _CustomModelBase.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        payload = {
            "custom_model_type": cls._model_type,  # type: ignore[attr-defined]
            "is_deployed": is_deployed,
            "order_by": order_by,
            "search_for": search_for,
        }
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls: Type[T_CustomModelBase], custom_model_id: str) -> T_CustomModelBase:
        """Get custom inference model by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            ID of the _CustomModelBase.

        Returns
        -------
        T_CustomModelBase
            Retrieved instance of _CustomModelBase

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = f"{cls._path}{custom_model_id}/"
        data = cls._client.get(path).json()
        if not cls._check_model_type(data):
            raise Exception(f"Requested model is not a {cls._model_type} model")  # type: ignore[attr-defined]
        return cls.from_server_data(data)

    def download_latest_version(self, file_path: str) -> None:
        """Download the latest custom inference model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            Path to create a file with custom model version content.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = f"{self._path}{self.id}/download/"

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    @classmethod
    def create(
        cls: Type[T_CustomModelBase],
        name: str,
        target_type: TARGET_TYPE,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> T_CustomModelBase:
        """Create a custom inference model.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str
            The name of the custom inference model.
        target_type: datarobot.TARGET_TYPE
            The target type of the custom inference model.
            Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
            `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.UNSTRUCTURED`,
            `datarobot.TARGET_TYPE.ANOMALY`]
        description: str, optional
            The description of the custom inference model.

        Returns
        -------
        CustomModelVersion
            Created instance of _CustomModelBase.

        Raises
        ------
            datarobot.errors.ClientError
                If the server responded with 4xx status.
            datarobot.errors.ServerError
                If the server responded with 5xx status.
        """
        payload = dict(
            custom_model_type=cls._model_type,  # type: ignore[attr-defined]
            name=name,
            description=description,
            **kwargs,
        )
        if target_type in CUSTOM_MODEL_TARGET_TYPE.ALL:
            payload["target_type"] = target_type

        # this will be removed when these params are fully deprecated
        if target_type == CUSTOM_MODEL_TARGET_TYPE.BINARY:
            payload["supports_binary_classification"] = True
        elif target_type == CUSTOM_MODEL_TARGET_TYPE.REGRESSION:
            payload["supports_regression"] = True
        elif target_type not in CUSTOM_MODEL_TARGET_TYPE.ALL:
            raise ClientError(
                "Unsupported target_type. target_type must be in  {}.".format(
                    CUSTOM_MODEL_TARGET_TYPE.ALL
                ),
                422,
            )

        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def copy_custom_model(
        cls: Type[T_CustomModelBase],
        custom_model_id: str,
    ) -> T_CustomModelBase:
        """Create a custom inference model by copying existing one.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            id of the custom inference model to copy

        Returns
        -------
        T_CustomModelBase
            Created a custom inference model.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = f"{cls._path}fromCustomModel/"
        response = cls._client.post(path, data={"custom_model_id": custom_model_id})
        return cls.from_server_data(response.json())

    def update(
        self, name: Optional[str] = None, description: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Update custom inference model properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str, optional
            New custom model name.
        description: str, optional
            New custom model description.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        payload = dict(name=name, description=description, **kwargs)
        url = f"{self._path}{self.id}/"
        response = self._client.patch(url, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))  # type: ignore[no-untyped-call]

    def refresh(self) -> None:
        """Update custom inference model with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.id)
        path = f"{url}{self.id}/"

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))  # type: ignore[no-untyped-call]

    def delete(self) -> None:
        """Delete custom inference model.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = f"{self._path}{self.id}/"
        self._client.delete(url)


class CustomInferenceModel(_CustomModelBase):
    """A custom inference model.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        The ID of the custom model.
    name: str
        The name of the custom model.
    language: str
        The programming language of the custom inference model.
        Can be "python", "r", "java" or "other".
    description: str
        The description of the custom inference model.
    target_type: datarobot.TARGET_TYPE
        Target type of the custom inference model.
        Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
        `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.UNSTRUCTURED`,
        `datarobot.TARGET_TYPE.ANOMALY`, `datarobot.TARGET_TYPE.TEXT_GENERATION`]
    target_name: str, optional
        Target feature name.
        It is optional(ignored if provided) for `datarobot.TARGET_TYPE.UNSTRUCTURED`
        or `datarobot.TARGET_TYPE.ANOMALY` target type.
    latest_version: datarobot.CustomModelVersion or None
        The latest version of the custom model if the model has a latest version.
    deployments_count: int
        Number of a deployments of the custom models.
    target_name: str
        The custom model target name.
    positive_class_label: str
        For binary classification projects, a label of a positive class.
    negative_class_label: str
        For binary classification projects, a label of a negative class.
    prediction_threshold: float
        For binary classification projects, a threshold used for predictions.
    training_data_assignment_in_progress: bool
        Flag describing if training data assignment is in progress.
    training_dataset_id: str, optional
        The ID of a dataset assigned to the custom model.
    training_dataset_version_id: str, optional
        The ID of a dataset version assigned to the custom model.
    training_data_file_name: str, optional
        The name of assigned training data file.
    training_data_partition_column: str, optional
        The name of a partition column in a training dataset assigned to the custom model.
    created_by: str
        The username of a user who created the custom model.
    updated_at: str
        ISO-8601 formatted timestamp of when the custom model was updated
    created_at: str
        ISO-8601 formatted timestamp of when the custom model was created
    network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
        Determines whether the given custom model is isolated, or can access the public network.
        Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
    maximum_memory: int, optional
        The maximum memory that might be allocated by the custom-model.
        If exceeded, the custom-model will be killed by k8s.
    replicas: int, optional
        A fixed number of replicas that will be deployed in the cluster
    is_training_data_for_versions_permanently_enabled: bool, optional
        Whether training data assignment on the version level is permanently enabled for the model.
    """

    _model_type: str = "inference"
    _converter = (
        _CustomModelBase._converter
        + {
            t.Key("language"): String(allow_blank=True),
            t.Key("target_name", optional=True): String(),
            t.Key("training_dataset_id", optional=True): String(),
            t.Key("training_dataset_version_id", optional=True): String(),
            t.Key("training_data_assignment_in_progress"): t.Bool(),
            t.Key("positive_class_label", optional=True): String(),
            t.Key("negative_class_label", optional=True): String(),
            t.Key("class_labels", optional=True): t.List(String()),
            t.Key("prediction_threshold", optional=True): t.Float(),
            t.Key("training_data_file_name", optional=True): String(),
            t.Key("training_data_partition_column", optional=True): String(),
            t.Key("network_egress_policy", optional=True): t.Enum(*NETWORK_EGRESS_POLICY.ALL),
            t.Key("maximum_memory", optional=True): Int(),
            t.Key("replicas", optional=True): Int(),
            t.Key(
                "is_training_data_for_versions_permanently_enabled", optional=True, default=False
            ): t.Bool(),
        }
    ).allow_extra("*")

    def _set_values(  # type: ignore[override] # pylint: disable=arguments-renamed
        self,
        language: str,
        training_data_assignment_in_progress: bool,
        is_training_data_for_versions_permanently_enabled: bool,
        target_name: Optional[str] = None,
        positive_class_label: Optional[str] = None,
        negative_class_label: Optional[str] = None,
        prediction_threshold: Optional[float] = None,
        class_labels: Optional[List[str]] = None,
        training_dataset_id: Optional[str] = None,
        training_dataset_version_id: Optional[str] = None,
        training_data_file_name: Optional[str] = None,
        training_data_partition_column: Optional[str] = None,
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY] = None,
        maximum_memory: Optional[int] = None,
        replicas: Optional[int] = None,
        **custom_model_kwargs: Any,
    ) -> None:
        super()._set_values(**custom_model_kwargs)

        self.language = language
        self.target_name = target_name
        self.training_dataset_id = training_dataset_id
        self.training_dataset_version_id = training_dataset_version_id
        self.training_data_assignment_in_progress = training_data_assignment_in_progress
        self.positive_class_label = positive_class_label
        self.negative_class_label = negative_class_label
        self.class_labels = class_labels
        self.prediction_threshold = prediction_threshold
        self.training_data_file_name = training_data_file_name
        self.training_data_partition_column = training_data_partition_column
        self.network_egress_policy = network_egress_policy
        self.maximum_memory = maximum_memory
        self.replicas = replicas
        self.is_training_data_for_versions_permanently_enabled = (
            is_training_data_for_versions_permanently_enabled
        )

    @classmethod
    def list(  # type: ignore[override] # pylint: disable=arguments-renamed
        cls,
        is_deployed: Optional[bool] = None,
        search_for: Optional[str] = None,
        order_by: Optional[str] = None,
    ) -> List[CustomInferenceModel]:
        """List custom inference models available to the user.

        .. versionadded:: v2.21

        Parameters
        ----------
        is_deployed: bool, optional
            Flag for filtering custom inference models.
            If set to `True`, only deployed custom inference models are returned.
            If set to `False`, only not deployed custom inference models are returned.
        search_for: str, optional
            String for filtering custom inference models - only custom
            inference models that contain the string in name or description will
            be returned.
            If not specified, all custom models will be returned
        order_by: str, optional
            Property to sort custom inference models by.
            Supported properties are "created" and "updated".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None which will result in
            custom models being returned in order of creation time descending.

        Returns
        -------
        List[CustomInferenceModel]
            A list of custom inference models.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status
        datarobot.errors.ServerError
            If the server responded with 5xx status
        """
        return super().list(is_deployed, order_by, search_for)

    @classmethod
    def get(cls, custom_model_id: str) -> CustomInferenceModel:
        """Get custom inference model by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom inference model.

        Returns
        -------
        CustomInferenceModel
            Retrieved custom inference model.

        Raises
        ------
        datarobot.errors.ClientError
            The ID the server responded with 4xx status.
        datarobot.errors.ServerError
            The ID the server responded with 5xx status.
        """
        return super().get(custom_model_id)

    # We must leave this method here in order for the docs to properly be generated
    # pylint: disable-next=useless-super-delegation
    def download_latest_version(self, file_path: str) -> None:
        """Download the latest custom inference model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            Path to create a file with custom model version content.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        super().download_latest_version(file_path)

    @classmethod
    def create(  # type: ignore[override] # pylint: disable=arguments-differ,arguments-renamed
        cls,
        name: str,
        target_type: TARGET_TYPE,
        target_name: Optional[str] = None,
        language: Optional[str] = None,
        description: Optional[str] = None,
        positive_class_label: Optional[str] = None,
        negative_class_label: Optional[str] = None,
        prediction_threshold: Optional[float] = None,
        class_labels: Optional[List[str]] = None,
        class_labels_file: Optional[str] = None,
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY] = None,
        maximum_memory: Optional[int] = None,
        replicas: Optional[int] = None,
        is_training_data_for_versions_permanently_enabled: Optional[bool] = None,
    ) -> CustomInferenceModel:
        """Create a custom inference model.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str
            Name of the custom inference model.
        target_type: datarobot.TARGET_TYPE
            Target type of the custom inference model.
            Values: [`datarobot.TARGET_TYPE.BINARY`, `datarobot.TARGET_TYPE.REGRESSION`,
            `datarobot.TARGET_TYPE.MULTICLASS`, `datarobot.TARGET_TYPE.UNSTRUCTURED`,
            `datarobot.TARGET_TYPE.TEXT_GENERATION`]
        target_name: str, optional
            Target feature name.
            It is optional(ignored if provided) for `datarobot.TARGET_TYPE.UNSTRUCTURED` target type.
        language: str, optional
            Programming language of the custom learning model.
        description: str, optional
            Description of the custom learning model.
        positive_class_label: str, optional
            Custom inference model positive class label for binary classification.
        negative_class_label: str, optional
            Custom inference model negative class label for binary classification.
        prediction_threshold: float, optional
            Custom inference model prediction threshold.
        class_labels: List[str], optional
            Custom inference model class labels for multiclass classification.
            Cannot be used with class_labels_file.
        class_labels_file: str, optional
            Path to file containing newline separated class labels for multiclass classification.
            Cannot be used with class_labels.
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`]
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s.
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster.
        is_training_data_for_versions_permanently_enabled: bool, optional
            Permanently enable training data assignment on the version level for the current model,
            instead of training data assignment on the model level.

        Returns
        -------
        CustomInferenceModel
            Created a custom inference model.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        if target_type in CUSTOM_MODEL_TARGET_TYPE.REQUIRES_TARGET_NAME and target_name is None:
            raise ValueError(
                f"target_name is required for custom models with target type {target_type}"
            )
        if class_labels and class_labels_file:
            raise ValueError("class_labels and class_labels_file cannot be used together")
        if class_labels_file:
            with open(class_labels_file) as f:  # pylint: disable=unspecified-encoding
                class_labels = [label for label in f.read().split("\n") if label]

        return super().create(
            name,
            target_type,
            description,
            language=language,
            target_name=target_name,
            positive_class_label=positive_class_label,
            negative_class_label=negative_class_label,
            prediction_threshold=prediction_threshold,
            class_labels=class_labels,
            network_egress_policy=network_egress_policy,
            maximum_memory=maximum_memory,
            replicas=replicas,
            is_training_data_for_versions_permanently_enabled=is_training_data_for_versions_permanently_enabled,
        )

    @classmethod
    def copy_custom_model(
        cls,
        custom_model_id: str,
    ) -> CustomInferenceModel:
        """Create a custom inference model by copying existing one.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom inference model to copy.

        Returns
        -------
        CustomInferenceModel
            Created a custom inference model.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        return super().copy_custom_model(custom_model_id)

    def update(  # type: ignore[override] # pylint: disable=arguments-differ,arguments-renamed
        self,
        name: Optional[str] = None,
        language: Optional[str] = None,
        description: Optional[str] = None,
        target_name: Optional[str] = None,
        positive_class_label: Optional[str] = None,
        negative_class_label: Optional[str] = None,
        prediction_threshold: Optional[float] = None,
        class_labels: Optional[List[str]] = None,
        class_labels_file: Optional[str] = None,
        is_training_data_for_versions_permanently_enabled: Optional[bool] = None,
    ) -> None:
        """Update custom inference model properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str, optional
            New custom inference model name.
        language: str, optional
            New custom inference model programming language.
        description: str, optional
            New custom inference model description.
        target_name: str, optional
            New custom inference model target name.
        positive_class_label: str, optional
            New custom inference model positive class label.
        negative_class_label: str, optional
            New custom inference model negative class label.
        prediction_threshold: float, optional
            New custom inference model prediction threshold.
        class_labels: List[str], optional
            custom inference model class labels for multiclass classification
            Cannot be used with class_labels_file
        class_labels_file: str, optional
            Path to file containing newline separated class labels for multiclass classification.
            Cannot be used with class_labels
        is_training_data_for_versions_permanently_enabled: bool, optional
            Permanently enable training data assignment on the version level for the current model,
            instead of training data assignment on the model level.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        if class_labels and class_labels_file:
            raise ValueError("class_labels and class_labels_file cannot be used together")
        if class_labels_file:
            with open(class_labels_file) as f:  # pylint: disable=unspecified-encoding
                class_labels = [label for label in f.read().split("\n") if label]

        super().update(
            name,
            description,
            language=language,
            target_name=target_name,
            positive_class_label=positive_class_label,
            negative_class_label=negative_class_label,
            prediction_threshold=prediction_threshold,
            class_labels=class_labels,
            is_training_data_for_versions_permanently_enabled=is_training_data_for_versions_permanently_enabled,
        )

    # We must leave this method here in order for the docs to properly be generated
    # pylint: disable-next=useless-super-delegation
    def refresh(self) -> None:
        """Update custom inference model with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        super().refresh()

    # We must leave this method here in order for the docs to properly be generated
    # pylint: disable-next=useless-super-delegation
    def delete(self) -> None:
        """Delete custom inference model.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        super().delete()

    @deprecated(
        deprecated_since_version="v3.2",
        will_remove_version="v3.5",
        message="Please use training data assignment on the model version level: "
        "CustomModelVersion.create_from_previous or CustomModelVersion.create_clean",
    )
    def assign_training_data(
        self,
        dataset_id: str,
        partition_column: Optional[str] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> None:
        """Assign training data to the custom inference model.

        .. versionadded:: v2.21

        Parameters
        ----------
        dataset_id: str
            The ID of the training dataset to be assigned.
        partition_column: str, optional
            The name of a partition column in the training dataset.
        max_wait: int, optional
            The max time to wait for a training data assignment.
            If set to None, then method will return without waiting.
            Defaults to 10 min.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status
        datarobot.errors.ServerError
            If the server responded with 5xx status
        """
        payload = {"dataset_id": dataset_id, "partition_column": partition_column}

        path = f"{self._path}{self.id}/trainingData/"

        response = self._client.patch(path, data=payload)

        if max_wait is not None:
            wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

        self.refresh()



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import String
from datarobot.models.confusion_chart import ConfusionChartTrafaret
from datarobot.utils.pagination import unpaginate

from ..api_object import APIObject
from .external_scores import DEFAULT_BATCH_SIZE


class ExternalConfusionChart(APIObject):
    """ Confusion chart for the model and prediction dataset with target.

    .. versionadded:: v2.21


    ``ClassMetrics`` is a dict containing the following:

        * ``class_name`` (string) name of the class
        * ``actual_count`` (int) number of times this class is seen in the validation data
        * ``predicted_count`` (int) number of times this class has been predicted for the \
          validation data
        * ``f1`` (float) F1 score
        * ``recall`` (float) recall score
        * ``precision`` (float) precision score
        * ``was_actual_percentages`` (list of dict) one vs all actual percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was predicted when is \
              was actually class (from 0 to 1)
        * ``was_predicted_percentages`` (list of dict) one vs all predicted percentages in format \
          specified below.
            * ``other_class_name`` (string) the name of the other class
            * ``percentage`` (float) the percentage of the times this class was actual predicted \
              (from 0 to 1)
        * ``confusion_matrix_one_vs_all`` (list of list) 2d list representing 2x2 one vs all matrix.
            * This represents the True/False Negative/Positive rates as integer for each class. \
              The data structure looks like:
            * ``[ [ True Negative, False Positive ], [ False Negative, True Positive ] ]``

    Attributes
    ----------
    dataset_id: str
        id of the external dataset with target
    model_id: str
        id of the model this confusion chart represents
    raw_data : dict
        All of the raw data for the Confusion Chart
    confusion_matrix : list of list
        The NxN confusion matrix
    classes : list
        The names of each of the classes
    class_metrics : list of dicts
        List of dicts with schema described as ``ClassMetrics`` above.

    """

    _path = "projects/{project_id}/models/{model_id}/datasetConfusionCharts/"
    _single_chart_path = (
        "projects/{project_id}/models/{model_id}/datasetConfusionCharts/{dataset_id}/"
    )
    _metadata_path = (
        "projects/{project_id}/models/{model_id}/datasetConfusionCharts/{dataset_id}/metadata/"
    )
    _converter = t.Dict(
        {t.Key("dataset_id"): String(), t.Key("data"): ConfusionChartTrafaret}
    ).ignore_extra("*")

    def __init__(self, dataset_id, data):
        self.dataset_id = dataset_id
        self.raw_data = data
        self.class_metrics = data["class_metrics"]
        self.confusion_matrix = data["confusion_matrix"]
        self.classes = data["classes"]

    def __repr__(self):
        return "ExternalConfusionChart(dataset_id={}, classes={})".format(
            self.dataset_id, self.classes
        )

    @classmethod
    def list(cls, project_id, model_id, dataset_id=None, offset=0, limit=100):
        """Retrieve list of the confusion charts for the model.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: str
            id of the model to retrieve a chart from
        dataset_id: str, optional
            if specified, only confusion chart for this dataset will be retrieved
        offset: int, optional
            this many results will be skipped, default: 0
        limit: int, optional
            at most this many results are returned, default: 100, max 1000.
            To return all results, specify 0

        Returns
        -------
            A list of :py:class:`ExternalConfusionChart <datarobot.ExternalConfusionChart>` objects
        """
        url = cls._path.format(project_id=project_id, model_id=model_id)
        params = {"limit": limit, "offset": offset}
        if dataset_id:
            params["datasetId"] = dataset_id
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            return [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)]
        r_data = cls._client.get(url, params=params).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, project_id, model_id, dataset_id):
        """Retrieve confusion chart for the model and external dataset.

        Parameters
        ----------
        project_id: str
            project id
        model_id: str
            model id
        dataset_id: str
            external dataset id with target

        Returns
        -------
            :py:class:`ExternalConfusionChart <datarobot.ExternalConfusionChart>` object

        """
        if dataset_id is None:
            raise ValueError("dataset_id must be specified")
        url = cls._single_chart_path.format(
            project_id=project_id, model_id=model_id, dataset_id=dataset_id
        )

        confusion_chart = cls._client.get(url).json()
        metadata = cls._get_metadata(
            project_id=project_id, model_id=model_id, dataset_id=dataset_id
        )
        confusion_chart["data"]["classes"] = metadata["classNames"]
        return cls.from_server_data(confusion_chart)

    @classmethod
    def _get_metadata(cls, project_id, model_id, dataset_id):
        """Retrieve confusion chart metadata.

        Parameters
        ----------
        project_id: str
            project id
        model_id: str
            model id
        dataset_id: str
            external dataset id with target

        Returns
        -------
            metadata : dict
                metadata of the confusion chart

        """
        url = cls._metadata_path.format(
            project_id=project_id, model_id=model_id, dataset_id=dataset_id
        )
        return cls._client.get(url).json()



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import cast, Iterable, List, Optional, Type, TYPE_CHECKING, TypeVar, Union

import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import DataStoreListTypes, DataStoreTypes
from datarobot.models.api_object import APIObject
from datarobot.models.dataset import Dataset
from datarobot.models.sharing import SharingAccess
from datarobot.utils.pagination import unpaginate

from ..utils import from_api, parse_time

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models.api_object import ServerDataType

    class DataSourceParametersPayload(TypedDict, total=False):
        data_store_id: Optional[str]
        catalog: Optional[str]
        table: Optional[str]
        schema: Optional[str]
        partition_column: Optional[str]
        query: Optional[str]
        fetch_size: Optional[int]
        path: Optional[str]


TDataSource = TypeVar("TDataSource", bound="DataSource")
TDataSourceParameters = TypeVar("TDataSourceParameters", bound="DataSourceParameters")

_data_source_params_converter = t.Dict(
    {
        t.Key("data_store_id"): t.Or(String(), t.Null),
        t.Key("catalog", optional=True): t.Or(String(), t.Null),
        t.Key("table", optional=True): t.Or(String(), t.Null),
        t.Key("schema", optional=True): t.Or(String(), t.Null),
        t.Key("partition_column", optional=True): t.Or(String(), t.Null),
        t.Key("query", optional=True): t.Or(String(), t.Null),
        t.Key("fetch_size", optional=True): t.Or(Int(), t.Null),
        t.Key("path", optional=True): t.Or(String(), t.Null),
    }
).ignore_extra("*")


class DataSourceParameters:
    """Data request configuration

    Attributes
    ----------
    data_store_id : str
        the id of the DataStore.
    table : str
        Optional. The name of specified database table.
    schema : str
        Optional. The name of the schema associated with the table.
    partition_column : str
        Optional. The name of the partition column.
    query : str
        Optional. The user specified SQL query.
    fetch_size : int
        Optional. A user specified fetch size in the range [1, 20000].
        By default a fetchSize will be assigned to balance throughput and memory usage
    path: str
        Optional. The user-specified path for BLOB storage
    """

    def __init__(
        self,
        data_store_id: Optional[str] = None,
        catalog: Optional[str] = None,
        table: Optional[str] = None,
        schema: Optional[str] = None,
        partition_column: Optional[str] = None,
        query: Optional[str] = None,
        fetch_size: Optional[int] = None,
        path: Optional[str] = None,
    ) -> None:
        _data_source_params_converter.check(
            {
                "data_store_id": data_store_id,
                "catalog": catalog,
                "table": table,
                "schema": schema,
                "partition_column": partition_column,
                "query": query,
                "fetch_size": fetch_size,
                "path": path,
            }
        )
        self.data_store_id = data_store_id
        self.catalog = catalog
        self.table = table
        self.schema = schema
        self.partition_column = partition_column
        self.query = query
        self.fetch_size = fetch_size
        self.path = path

    def collect_payload(self) -> DataSourceParametersPayload:
        return {
            "data_store_id": self.data_store_id,
            "catalog": self.catalog,
            "table": self.table,
            "schema": self.schema,
            "partition_column": self.partition_column,
            "query": self.query,
            "fetch_size": self.fetch_size,
            "path": self.path,
        }

    @classmethod
    def from_server_data(cls, data: ServerDataType) -> DataSourceParameters:
        converted_data = _data_source_params_converter.check(from_api(data))
        return cls(**converted_data)

    def __eq__(self, other: TDataSourceParameters) -> bool:  # type: ignore[override]
        self_payload = self.collect_payload()
        other_payload = other.collect_payload()
        del self_payload["data_store_id"]
        del other_payload["data_store_id"]
        return self_payload == other_payload


class DataSource(APIObject):
    """A data source. Represents data request

    Attributes
    ----------
    id : str
        the id of the data source.
    type : str
        the type of data source.
    canonical_name : str
        the user-friendly name of the data source.
    creator : str
        the id of the user who created the data source.
    updated : datetime.datetime
        the time of the last update.
    params : DataSourceParameters
        a list specifying data source parameters.
    role : str or None
        if a string, represents a particular level of access and should be one of
        ``datarobot.enums.SHARING_ROLE``.  For more information on the specific access levels, see
        the :ref:`sharing <sharing>` documentation.  If None, can be passed to a `share`
        function to revoke access for a specific user.
    """

    _path = "externalDataSources/"
    _converter = t.Dict(
        {
            t.Key("id", optional=True) >> "data_source_id": String(),
            t.Key("type") >> "data_source_type": String(),
            t.Key("canonical_name"): String(),
            t.Key("creator"): String(),
            t.Key("params"): _data_source_params_converter,
            t.Key("updated"): parse_time,
            t.Key("role"): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        data_source_id: Optional[str] = None,
        data_source_type: Optional[str] = None,
        canonical_name: Optional[str] = None,
        creator: Optional[str] = None,
        updated: Optional[datetime] = None,
        params: Optional[DataSourceParameters] = None,
        role: Optional[str] = None,
    ) -> None:
        self._id = data_source_id
        self._type = data_source_type
        self.canonical_name = canonical_name
        self._creator = creator
        self._updated = updated
        self.params = params
        self.role = role

    @classmethod
    def list(cls, typ: Optional[DataStoreListTypes] = None) -> List[DataSource]:
        """
        Returns list of available data sources.

        Parameters
        ----------
        typ : DataStoreListTypes
            If specified, filters by specified datasource type. If not specified it will
            default to DataStoreListTypes.DATABASES

        Returns
        -------
        data_sources : list of DataSource instances
            contains a list of available data sources.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_sources = dr.DataSource.list()
            >>> data_sources
            [DataSource('Diagnostics'), DataSource('Airlines 100mb'), DataSource('Airlines 10mb')]
        """
        if typ is not None:
            r_data = cls._client.get(cls._path, params={"type": str(typ)}).json()
        else:
            r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls: Type[TDataSource], data_source_id: str) -> TDataSource:
        """
        Gets the data source.

        Parameters
        ----------
        data_source_id : str
            the identifier of the data source.

        Returns
        -------
        data_source : DataSource
            the requested data source.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_source = dr.DataSource.get('5a8ac9ab07a57a0001be501f')
            >>> data_source
            DataSource('Diagnostics')
        """
        return cls.from_location(f"{cls._path}{data_source_id}/")

    @classmethod
    def create(
        cls: Type[TDataSource],
        data_source_type: Union[str, DataStoreTypes],
        canonical_name: str,
        params: DataSourceParameters,
    ) -> TDataSource:
        """
        Creates the data source.

        Parameters
        ----------
        data_source_type : str or DataStoreTypes
            the type of data source.
        canonical_name : str
            the user-friendly name of the data source.
        params : DataSourceParameters
            a list specifying data source parameters.

        Returns
        -------
        data_source : DataSource
            the created data source.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> params = dr.DataSourceParameters(
            ...     data_store_id='5a8ac90b07a57a0001be501e',
            ...     query='SELECT * FROM airlines10mb WHERE "Year" >= 1995;'
            ... )
            >>> data_source = dr.DataSource.create(
            ...     data_source_type='jdbc',
            ...     canonical_name='airlines stats after 1995',
            ...     params=params
            ... )
            >>> data_source
            DataSource('airlines stats after 1995')
        """
        payload = {
            "type": str(data_source_type),
            "canonicalName": canonical_name,
            "params": params.collect_payload(),
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self, canonical_name: Optional[str] = None, params: Optional[DataSourceParameters] = None
    ) -> None:
        """
        Creates the data source.

        Parameters
        ----------
        canonical_name : str
            optional, the user-friendly name of the data source.
        params : DataSourceParameters
            optional, the identifier of the DataDriver.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_source = dr.DataSource.get('5ad840cc613b480001570953')
            >>> data_source
            DataSource('airlines stats after 1995')
            >>> params = dr.DataSourceParameters(
            ...     query='SELECT * FROM airlines10mb WHERE "Year" >= 1990;'
            ... )
            >>> data_source.update(
            ...     canonical_name='airlines stats after 1990',
            ...     params=params
            ... )
            >>> data_source
            DataSource('airlines stats after 1990')
        """
        if params is None and self.params is None:
            raise ValueError("Must supply 'params' parameter.")
        payload = {
            "canonicalName": canonical_name or self.canonical_name,
            "params": params.collect_payload()
            if params is not None
            else cast(DataSourceParameters, self.params).collect_payload(),
        }
        r_data = self._client.patch(f"{self._path}{self.id}/", data=payload).json()
        self.canonical_name = r_data["canonicalName"]
        self.params = DataSourceParameters.from_server_data(r_data.pop("params"))

    def delete(self) -> None:
        """Removes the DataSource"""
        self._client.delete(f"{self._path}{self.id}/")

    @classmethod
    def from_server_data(
        cls: Type[TDataSource],
        data: ServerDataType,
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> TDataSource:
        converted_data = cls._converter.check(from_api(data))
        params = converted_data.pop("params")
        data_store_id = params.pop("data_store_id")
        converted_data["params"] = DataSourceParameters(data_store_id, **params)
        return cls(**converted_data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.canonical_name or self.id}')"

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def creator(self) -> Optional[str]:
        return self._creator

    @property
    def type(self) -> Optional[str]:
        return self._type

    @property
    def updated(self) -> Optional[datetime]:
        return self._updated

    def get_access_list(self) -> List[SharingAccess]:
        """Retrieve what users have access to this data source

        .. versionadded:: v2.14

        Returns
        -------
        list of :class:`SharingAccess <datarobot.SharingAccess>`
        """
        url = f"{self._path}{self.id}/accessControl/"
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, {}, self._client)
        ]

    def share(self, access_list: List[SharingAccess]) -> None:
        """Modify the ability of users to access this data source

        .. versionadded:: v2.14

        Parameters
        ----------
        access_list: list of :class:`SharingAccess <datarobot.SharingAccess>`
            The modifications to make.

        Raises
        ------
        datarobot.ClientError:
            If you do not have permission to share this data source, if the user you're sharing with
            doesn't exist, if the same user appears multiple times in the access_list, or if these
            changes would leave the data source without an owner.

        Examples
        --------
        Transfer access to the data source from old_user@datarobot.com to new_user@datarobot.com

        .. code-block:: python

            from datarobot.enums import SHARING_ROLE
            from datarobot.models.data_source import DataSource
            from datarobot.models.sharing import SharingAccess

            new_access = SharingAccess(
                "new_user@datarobot.com",
                SHARING_ROLE.OWNER,
                can_share=True,
            )
            access_list = [
                SharingAccess("old_user@datarobot.com", SHARING_ROLE.OWNER, can_share=True),
                new_access,
            ]

            DataSource.get('my-data-source-id').share(access_list)
        """
        payload = {"data": [access.collect_payload() for access in access_list]}
        self._client.patch(
            f"{self._path}{self.id}/accessControl/", data=payload, keep_attrs={"role"}
        )

    def create_dataset(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        do_snapshot: Optional[bool] = None,
        persist_data_after_ingestion: Optional[bool] = None,
        categories: Optional[List[str]] = None,
        credential_id: Optional[str] = None,
        use_kerberos: Optional[bool] = None,
    ) -> Dataset:
        """
        Create a :py:class:`Dataset <datarobot.models.Dataset>` from this data source.

        .. versionadded:: v2.22

        Parameters
        ----------
        username: string, optional
            The username for database authentication.
        password: string, optional
            The password (in cleartext) for database authentication. The password
            will be encrypted on the server side in scope of HTTP request and never saved or stored.
        do_snapshot: bool, optional
            If unset, uses the server default: True.
            If true, creates a snapshot dataset; if
            false, creates a remote dataset. Creating snapshots from non-file sources requires an
            additional permission, `Enable Create Snapshot Data Source`.
        persist_data_after_ingestion: bool, optional
            If unset, uses the server default: True.
            If true, will enforce saving all data
            (for download and sampling) and will allow a user to view extended data profile
            (which includes data statistics like min/max/median/mean, histogram, etc.). If false,
            will not enforce saving data. The data schema (feature names and types) still will be
            available. Specifying this parameter to false and `doSnapshot` to true will result in
            an error.
        categories: list[string], optional
            An array of strings describing the intended use of the dataset. The
            current supported options are "TRAINING" and "PREDICTION".
        credential_id: string, optional
            The ID of the set of credentials to
            use instead of user and password. Note that with this change, username and password
            will become optional.
        use_kerberos: bool, optional
            If unset, uses the server default: False.
            If true, use kerberos authentication for database authentication.

        Returns
        -------
        response: Dataset
            The Dataset created from the uploaded data
        """
        if not self.id:
            raise ValueError("Missing required DataSource ID - needed to create Dataset.")
        return Dataset.create_from_data_source(
            self.id,
            username=username,
            password=password,
            do_snapshot=do_snapshot,
            persist_data_after_ingestion=persist_data_after_ingestion,
            categories=categories,
            credential_id=credential_id,
            use_kerberos=use_kerberos,
        )



#
# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import webbrowser

from datarobot.utils.logger import get_logger

logger = get_logger(__name__)


class BrowserMixin:
    """A mixin to allow opening a class' relevant URI in a web browser.

    Class must implement get_uri()
    """

    def get_uri(self) -> str:
        raise NotImplementedError

    def open_in_browser(self) -> None:
        """
        Opens class' relevant web browser location.
        If default browser is not available the URL is logged.

        Note:
        If text-mode browsers are used, the calling process will block
        until the user exits the browser.
        """
        try:
            # Returns default browser - so if one does not exist or is not set an exception is raised
            webbrowser.get()
            webbrowser.open(self.get_uri())
        except webbrowser.Error:
            logger.info("Please visit: %s", self.get_uri())



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Generator, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from datarobot.rest import RESTClientObject


def unpaginate(
    initial_url: str,
    initial_params: Union[None, Dict[Any, Any]],
    client: RESTClientObject,
) -> Generator[Any, None, None]:
    """Iterate over a paginated endpoint and get all results

    Assumes the endpoint follows the "standard" pagination interface (data stored under "data",
    "next" used to link next page, "offset" and "limit" accepted as query parameters).

    Yields
    ------
    data : dict
        a series of objects from the endpoint's data, as raw server data
    """
    resp_data = client.get(initial_url, params=initial_params).json()
    yield from resp_data["data"]
    while resp_data["next"] is not None:
        next_url = resp_data["next"]
        resp_data = client.get(next_url).json()
        yield from resp_data["data"]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import io
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np

from datarobot.enums import (
    DEFAULT_VISUAL_AI_FORCE_SIZE,
    DEFAULT_VISUAL_AI_IMAGE_FORMAT,
    DEFAULT_VISUAL_AI_IMAGE_QUALITY,
    DEFAULT_VISUAL_AI_IMAGE_QUALITY_KEEP_IF_POSSIBLE,
    DEFAULT_VISUAL_AI_IMAGE_RESAMPLE_METHOD,
    DEFAULT_VISUAL_AI_IMAGE_SIZE,
    DEFAULT_VISUAL_AI_IMAGE_SUBSAMPLING,
    DEFAULT_VISUAL_AI_SHOULD_RESIZE,
    SUPPORTED_IMAGE_FORMATS,
)

try:
    from PIL import Image
except ImportError:
    msg = (
        "Image transformation operations require installation of datarobot library, "
        "with optional `images` dependency. To install library with image support"
        "please use `pip install datarobot[images]`"
    )
    raise ImportError(msg)


class ImageOptions:
    """
    Image options class. Class holds image options related to image resizing and image reformatting.

    should_resize: bool
        Whether input image should be resized to new dimensions.
    force_size: bool
        Whether the image size should fully match the new requested size. If the original
        and new image sizes have different aspect ratios, specifying True will force a resize
        to exactly match the requested size. This may break the aspect ratio of the original
        image. If False, the resize method modifies the image to contain a thumbnail version
        of itself, no larger than the given size, that preserves the image's aspect ratio.
    image_size: Tuple[int, int]
        New image size (width, height). Both values (width, height) should be specified and contain
        a positive value. Depending on the value of `force_size`, the image will be resized exactly
        to the given image size or will be resized into a thumbnail version of itself, no larger
        than the given size.
    image_format: ImageFormat | str
        What image format will be used to save result image after transformations. For example
        (ImageFormat.JPEG, ImageFormat.PNG). Values supported are in line with values supported
        by DataRobot. If no format is specified by passing `None` value original image format
        will be preserved.
    image_quality: int or None
        The image quality used when saving image. When None is specified, a value will
        not be passed and Pillow library will use its default.
    resample_method: ImageResampleMethod
        What resampling method should be used when resizing image.
    keep_quality: bool
        Whether the image quality is kept (when possible). If True, for JPEG images quality will
        be preserved. For other types, the value specified in `image_quality` will be used.
    """

    def __init__(
        self,
        should_resize: bool = DEFAULT_VISUAL_AI_SHOULD_RESIZE,
        force_size: bool = DEFAULT_VISUAL_AI_FORCE_SIZE,
        image_size: Tuple[int, int] = DEFAULT_VISUAL_AI_IMAGE_SIZE,
        image_format: Optional[
            str
        ] = DEFAULT_VISUAL_AI_IMAGE_FORMAT,  # This constant is equal to None
        image_quality: int = DEFAULT_VISUAL_AI_IMAGE_QUALITY,
        image_subsampling: Optional[
            int
        ] = DEFAULT_VISUAL_AI_IMAGE_SUBSAMPLING,  # This constant is equal to None
        resample_method: int = DEFAULT_VISUAL_AI_IMAGE_RESAMPLE_METHOD,
        keep_quality: bool = DEFAULT_VISUAL_AI_IMAGE_QUALITY_KEEP_IF_POSSIBLE,
    ) -> None:
        self.should_resize = should_resize
        self.force_size = force_size
        self.image_size = image_size
        self.resample_method = resample_method
        self.image_quality = image_quality
        self.image_format = image_format
        self.image_quality = image_quality
        self.image_subsampling = image_subsampling
        self.keep_quality = keep_quality
        self._validate()

    def _validate(self) -> None:  # pylint: disable=missing-function-docstring
        if self.should_resize is None:
            raise ValueError("Image transformation requires value `should_resize` parameter.")
        elif self.should_resize is True:
            if not self.image_size:
                raise ValueError("When resizing image `image_size` value is required.")
            width, height = self.image_size
            if width is None or width <= 0:
                raise ValueError("Image size width value should be positive number.")
            if height is None or height <= 0:
                raise ValueError("Image size height value should be positive number.")
        if self.image_format and self.image_format not in SUPPORTED_IMAGE_FORMATS:
            raise ValueError(
                "Invalid image_format value. Please specify `None` "
                "to preserve current image format or one of supported "
                "image formats: {}".format(SUPPORTED_IMAGE_FORMATS)
            )


def get_image_save_kwargs(image: Image, image_options: ImageOptions) -> Dict[str, Any]:
    """Return pillow save args depending on format and image options."""
    image_format = image_options.image_format if image_options.image_format else image.format
    kwargs = dict(compression="None", format=image_format)
    # add optional save parameters format dependant
    if image_format in {"JPEG", "PNG"}:
        kwargs["optimize"] = True
    if image_format == "JPEG":
        # keep can work only when source and target are jpeg
        if image_format == image.format and image_options.keep_quality:
            kwargs["quality"] = "keep"
        elif image_options.image_quality is not None:
            kwargs["quality"] = image_options.image_quality
    if image_options.image_subsampling is not None:
        kwargs["subsampling"] = image_options.image_subsampling
    return kwargs


def get_bytes_from_image(image: Image, image_options: ImageOptions) -> Optional[bytes]:
    """
    Save PIL image with in specified format and with specified options and return image image_bytes.

    Parameters
    ----------
    image: PIL.Image
        Image object instance
    image_options: class ImageOptions
        Instance of class with image options

    Returns
    -------
    image_bytes representing Image
    """
    if image:
        bytes = io.BytesIO()
        image.save(fp=bytes, **get_image_save_kwargs(image, image_options))
        return bytes.getvalue()
    return None


def get_image_from_bytes(image_bytes: Union[bytes, io.BytesIO]) -> Image:
    """
    Create PIL Image instance using input bytes.

    Parameters
    ----------
    image_bytes: bytes | BytesIO
        Image in a form of bytes.

    Returns
    -------
    PIL.Image instance
    """
    if isinstance(image_bytes, bytes):
        image_bytes = io.BytesIO(image_bytes)
    image = Image.open(image_bytes)
    return image


def _scale_image_to_8bit_range(image: Image) -> Image:
    """
    This prevents a known bug of Pillow. When converting images, PIL clips the values.
    Thus, we have to manually scale the values, so they fit into 8 bit.
    https://github.com/python-pillow/Pillow/issues/2574
    """
    im_data = np.array(image)
    im_data = im_data * (2**8 / im_data.max())
    image = Image.fromarray(im_data.astype(np.int32))
    return image


def format_image_bytes(image_bytes: bytes, image_options: ImageOptions) -> Optional[bytes]:
    """
    Format input image image_bytes and return reformatted image also in a form of image_bytes.

    Parameters
    ----------
    image_bytes: bytes
        Image in a form of image_bytes
    image_options : ImageOptions
        Class holding image formatting and image resize parameters

    Returns
    -------
    Image in a form of image_bytes
    """
    image = get_image_from_bytes(image_bytes)
    image_format = image.format
    # if image should be resized calculate new image size and resize original image
    if image_options.should_resize:
        if image_options.force_size:
            image = image.resize(
                size=image_options.image_size, resample=image_options.resample_method
            )
        else:
            image.thumbnail(size=image_options.image_size, resample=image_options.resample_method)

    if image.mode != "RGB":
        if image.mode == "I":
            image = _scale_image_to_8bit_range(image)
        image = image.convert("RGB")

    # Some PIL operations loose the image_format information.
    image.format = image_format

    # convert to target image format and return as image_bytes
    return get_bytes_from_image(image, image_options)
